/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dgws;

import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.AuthenticationLevel;
import dk.sosi.seal.model.CareProvider;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SecurityTokenRequest;
import dk.sosi.seal.model.SecurityTokenResponse;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.model.SystemIDCard;
import dk.sosi.seal.model.constants.SubjectIdentifierTypeValues;
import dk.sosi.seal.pki.SOSIFederation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;
import org.apache.log4j.Logger;
import org.w3c.dom.Document;

import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Date;
import java.util.Properties;

/**
 * Factory for DGWS requests, which caches a SystemIDCard instance such that we don't need to request a new one from
 * STS for each service invocation.
 */
public class DGWSRequestFactory {

    private static final Logger log = Logger.getLogger(DGWSRequestFactory.class);

    private final STSConfig stsConfig;
    private final HttpClientService httpClientService;
    private final SOSIFactory sosiFactory;

    private SystemIDCard stsSignedIdCard;

    public DGWSRequestFactory(STSConfig stsConfig, HttpClientService httpClientService) {
        this.stsConfig = stsConfig;
        this.httpClientService = httpClientService;

        Properties properties = SignatureUtil.setupCryptoProviderForJVM();
        properties.setProperty(SOSIFactory.PROPERTYNAME_SOSI_VALIDATE, Boolean.toString(true));
        CredentialVault vault = new ClasspathCredentialVault(properties, stsConfig.getKeystorePath(), stsConfig.getKeystorePassword());
        if (Boolean.parseBoolean(System.getenv("NSP_TEST_FEDERATION")) || Boolean.getBoolean(DGWSRequestFactory.class.getName() + ".test")) {
            log.warn("Using NSP Test Federation");
            sosiFactory = new SOSIFactory(new SOSITestFederation(new Properties()), vault, properties);
        } else {
            log.info("Using NSP Production Federation");
            sosiFactory = new SOSIFactory(new SOSIFederation(new Properties()), vault, properties);
        }
    }

    public void ensureValidIdCard() {
        if (stsSignedIdCard != null) {
            if (isCurrentIdCardValid()) {
                // We have a valid idcard with at least 5 hours remaining.
                return;
            } else {
                log.info("Renewing IDCard from " + stsConfig.getStsEndpoint());
            }
        } else {
            log.info("Getting first IDCard from " + stsConfig.getStsEndpoint());
        }

        Document securityTokenRequest = buildSecurityTokenRequest();
        String requestXml = XmlUtil.node2String(securityTokenRequest, false, true);

        String responseXml = httpClientService.soapRequest(stsConfig.getStsEndpoint(), "\"Issue\"", requestXml);

        stsSignedIdCard = buildSystemIDCard(responseXml);

        log.info("Got new IDCard based on '" + stsSignedIdCard.getAlternativeIdentifier() + "' issued by " + stsSignedIdCard.getIssuer());
    }

    private boolean isCurrentIdCardValid() {
        Date expiryDate = stsSignedIdCard.getExpiryDate();
        return expiryDate.toInstant().minus(5, ChronoUnit.HOURS).isAfter(Instant.now());
    }

    public Request createNewRequest(String flowId) {
        Request sosiRequest = sosiFactory.createNewRequest(false, flowId);
        sosiRequest.setIDCard(stsSignedIdCard);
        sosiRequest.setDGWSVersion("1.0.1");
        return sosiRequest;
    }


    private SystemIDCard buildSystemIDCard(String responseXml) {
        SecurityTokenResponse securityTokenResponse = sosiFactory.deserializeSecurityTokenResponse(responseXml);
        SystemIDCard signedIdCard = (SystemIDCard) securityTokenResponse.getIDCard();

        if (securityTokenResponse.isFault() || signedIdCard == null) {
            log.error("Got a SOAP Fault from STS. The detailed message can be seen in the xml response");
            if (securityTokenResponse.isFault()) {
                log.error("A SOAP Fault was received: " + securityTokenResponse.getFaultString());
            }
            throw new IllegalStateException("SOAP Fault from STS");
        }
        return signedIdCard;
    }

    private Document buildSecurityTokenRequest() {
        CareProvider careProvider = new CareProvider(SubjectIdentifierTypeValues.CVR_NUMBER, stsConfig.getOrganizationCode(), stsConfig.getOrganizationName());
        X509Certificate certificate = sosiFactory.getCredentialVault().getSystemCredentialPair().getCertificate();
        SystemIDCard systemIDCard = sosiFactory.createNewSystemIDCard(stsConfig.getSystemName(), careProvider, AuthenticationLevel.VOCES_TRUSTED_SYSTEM, null, null, certificate, null);

        SecurityTokenRequest securityTokenRequest = sosiFactory.createNewSecurityTokenRequest();
        securityTokenRequest.setIDCard(systemIDCard);
        return securityTokenRequest.serialize2DOMDocument();
    }
}
