/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dgws;

import dk.sosi.seal.model.Request;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpException;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.UUID;
import java.util.stream.Collectors;

import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@CbsTest
@SuppressWarnings("java:S5979") /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
class DGWSRequestFactoryTest {

    private static String stsStubResponse;
    private static String stsFaultResponse;
    private static String stsNoIdCardResponse;

    @Mock
    private HttpClientService httpClientService;

    private DGWSRequestFactory dgwsRequestFactory;

    @BeforeAll
    static void beforeAll() {
        stsStubResponse = readFile("sts-stub-response.xml");
        stsFaultResponse = readFile("sts-fault-response.xml");
        stsNoIdCardResponse = readFile("sts-no-id-card-response.xml");
    }

    @BeforeEach
    void setUp() {
        STSConfig stsConfig = new STSConfig(
            "http://test1.ekstern-test.nspop.dk:8080/sts/services/NewSecurityTokenService",
            "Sundhedsdatastyrelsen",
            "33257872",
            "Sundhedsdatastyrelsen - COVID-19 Borgerstatus.jks",
            "Test1234",
            "CBS"
        );

        System.setProperty(DGWSRequestFactory.class.getName() + ".test", "true");

        dgwsRequestFactory = new DGWSRequestFactory(stsConfig, httpClientService);
    }

    @Test
    void should_get_request_with_id_card() {
        when(httpClientService.soapRequest(any(), any(), any())).thenReturn(stsStubResponse);

        dgwsRequestFactory.ensureValidIdCard();

        Request newRequest = dgwsRequestFactory.createNewRequest(UUID.randomUUID().toString());

        assertNotNull(newRequest.getIDCard());
    }

    @Test
    void should_throw_if_sts_request_fails() {
        when(httpClientService.soapRequest(any(), any(), any())).thenThrow(new HttpException("STS Error"));

        assertThrows(HttpException.class, () -> dgwsRequestFactory.ensureValidIdCard());
    }

    @Test
    void should_throw_if_sts_returns_fault() {
        when(httpClientService.soapRequest(any(), any(), any())).thenReturn(stsFaultResponse);
        assertThrows(IllegalStateException.class, () -> dgwsRequestFactory.ensureValidIdCard());
    }

    @Test
    void should_throw_if_sts_response_has_no_id_card() {
        when(httpClientService.soapRequest(any(), any(), any())).thenReturn(stsNoIdCardResponse);
        assertThrows(IllegalStateException.class, () -> dgwsRequestFactory.ensureValidIdCard());
    }

    @Test
    void should_refresh_if_id_card_is_expired() {
        // Stub response has expired id card
        when(httpClientService.soapRequest(any(), any(), any())).thenReturn(stsStubResponse);

        //set expired id card as current
        dgwsRequestFactory.ensureValidIdCard();

        // verify that first id card is requested from STS
        verify(httpClientService).soapRequest(any(), any(), any());

        Mockito.reset(httpClientService);
        when(httpClientService.soapRequest(any(), any(), any())).thenReturn(stsStubResponse);

        // This call should try to renew the id card
        dgwsRequestFactory.ensureValidIdCard();
        verify(httpClientService).soapRequest(any(), any(), any());
    }

    private static String readFile(String path) {
        try (InputStream is = DGWSRequestFactoryTest.class.getClassLoader().getResourceAsStream(path)) {
            if (is == null) return null;
            try (InputStreamReader isr = new InputStreamReader(is, StandardCharsets.UTF_8);
                 BufferedReader reader = new BufferedReader(isr)) {
                return reader.lines().collect(Collectors.joining(System.lineSeparator()));
            }
        } catch (IOException e) {
            return null;
        }
    }
}
