/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.kafka;

import dk.sds.nsp.kafka.clients.consumer.NSPKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import org.apache.kafka.clients.consumer.Consumer;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.apache.kafka.clients.consumer.ConsumerRecords;
import org.apache.kafka.common.errors.WakeupException;
import org.apache.log4j.Logger;

import javax.annotation.PreDestroy;
import java.io.Closeable;
import java.util.Collections;
import java.util.ConcurrentModificationException;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

public class CbsKafkaConsumer implements Closeable {
    private static final Logger log = Logger.getLogger(CbsKafkaConsumer.class);
    private final AtomicBoolean closed = new AtomicBoolean(false);
    private final Consumer<String, DataFetcherJob> consumer;
    private final String clientId;
    private final long pollTimeout;

    public CbsKafkaConsumer(String topic, String clientId, NSPKafkaConsumer<String, DataFetcherJob> consumer, long pollTimeout) {
        this.clientId = clientId;
        this.consumer = consumer;
        this.consumer.subscribe(Collections.singletonList(topic));
        this.pollTimeout = pollTimeout;
    }

    /**
     * Shutdown handling per KafkaConsumer JavaDoc section "Multi-threaded Processing"
     *
     * @see org.apache.kafka.clients.consumer.KafkaConsumer
     */
    @PreDestroy
    @Override
    public void close() {
        if (!closed.getAndSet(true)) {
            log.info("Closing " + clientId);
            try {
                //try and close consumer (throws ConcurrentModificationException if we are not on the thread currently using the consumer)
                consumer.close(1, TimeUnit.MINUTES);
            } catch (ConcurrentModificationException e) {
                consumer.wakeup(); //generate a WakeupException on the thread currently using the consumer (which either aborts current poll or causes next call to poll to cause WakeupException)
            }
        }
    }

    public synchronized Set<DataFetcherJob> pollDataFetcherJobs() {
        ConsumerRecords<String, DataFetcherJob> consumerRecords;
        try {
            consumerRecords = consumer.poll(pollTimeout);
        } catch (WakeupException e) {
            log.debug("WakeupException was raised on " + clientId);
            if (!closed.get()) throw e; //We are not closing. Propagate the exception without closing the consumer
            consumer.close(1, TimeUnit.MINUTES);
            return Collections.emptySet();
        }
        log.info(clientId + " received " + consumerRecords.count() + " record(s) from Kafka");

        if (log.isTraceEnabled()) {
            for (ConsumerRecord<String, DataFetcherJob> record : consumerRecords) {
                log.trace("---------\n" +
                    clientId + " received record:\n" +
                    "\t Value:" + record.value() + "\n" +
                    "\t Topic:" + record.topic() + "\n" +
                    "\t Partition:" + record.partition() + "\n" +
                    "\t Offset:" + record.offset());
            }
        }

        Set<DataFetcherJob> jobs = new HashSet<>(consumerRecords.count());
        consumerRecords.forEach(record -> jobs.add(record.value()));
        return jobs;
    }

    public int getNumberOfTopicPartitions(String topic) {
        return consumer.partitionsFor(topic).size();
    }

}
