/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.nas.facade;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.medcom.dgws._2006._04.dgws_1_0.Linking;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.oasis_open.docs.wsn.b_2.CreatePullPointResponse;
import org.oasis_open.docs.wsn.b_2.GetMessagesResponse;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.SubscribeResponse;
import org.oasis_open.docs.wsn.bw_2.*;

import javax.xml.ws.wsaddressing.W3CEndpointReference;
import javax.xml.ws.wsaddressing.W3CEndpointReferenceBuilder;
import java.util.List;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

@CbsTest
@SuppressWarnings("java:S5979") /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
class NasClientImplTest {

    @Mock
    private DGWSHeaderProvider dgwsHeaderProvider;

    @Mock
    private NotificationProducer notificationProducer;

    @Mock
    private CreatePullPoint createPullPoint;

    @Mock
    private PullPointFactory pullPointFactory;

    @Mock
    private PullPoint pullPoint;

    private NasClientImpl nasClient;

    @BeforeEach
    void setUp() {
        when(dgwsHeaderProvider.getHeaders()).thenReturn(new DGWSHeaders(null, getHeader()));

        nasClient = new NasClientImpl(dgwsHeaderProvider, notificationProducer, createPullPoint, pullPointFactory);
    }


    @Test
    void should_return_pull_point_reference_when_creating_pull_point() throws Exception {
        CreatePullPointResponse pullPointResponse = new CreatePullPointResponse();
        W3CEndpointReference endpointReference = getEndpointReference("http://pull.point");
        pullPointResponse.setPullPoint(endpointReference);
        when(createPullPoint.createPullPoint(any(), any(), any())).thenReturn(pullPointResponse);

        W3CEndpointReference pullPoint = nasClient.createPullPoint();

        assertNotNull(pullPoint);
    }

    @Test
    void should_fail_if_unable_to_create_pull_point() throws Exception {
        when(createPullPoint.createPullPoint(any(), any(), any())).thenThrow(new UnableToCreatePullPointFault());

        assertThrows(NasException.class, () -> nasClient.createPullPoint());
    }

    @Test
    void should_return_subscription_reference_when_subscribing() throws Exception {
        SubscribeResponse subscribeResponse = new SubscribeResponse();
        subscribeResponse.setSubscriptionReference(getEndpointReference("http://subscribe.now"));
        when(notificationProducer.subscribe(any(), any(), any())).thenReturn(subscribeResponse);

        W3CEndpointReference subscription = nasClient.subscribe(getEndpointReference("http://pull.point"), "topic");

        assertNotNull(subscription);
    }

    @Test
    void should_fail_if_unable_to_subscribe() throws Exception {
        when(notificationProducer.subscribe(any(), any(), any())).thenThrow(new SubscribeCreationFailedFault());

        W3CEndpointReference endpointReference = getEndpointReference("http://pull.point");

        assertThrows(NasException.class, () -> nasClient.subscribe(endpointReference, "topic"));
    }

    @Test
    void should_return_messages_on_request() throws Exception {

        when(pullPointFactory.getPullPoint(any())).thenReturn(pullPoint);
        GetMessagesResponse getMessagesResponse = new GetMessagesResponse();
        getMessagesResponse.getNotificationMessage().add(new NotificationMessageHolderType());
        when(pullPoint.getMessages(any(), any(), any())).thenReturn(getMessagesResponse);

        List<NotificationMessageHolderType> messages = nasClient.getMessages(getEndpointReference("http://pull.point"), 1000L, UUID.randomUUID());
        assertFalse(messages.isEmpty());
    }

    @Test
    void should_fail_if_unable_to_get_messages() throws Exception {
        when(pullPointFactory.getPullPoint(any())).thenReturn(pullPoint);

        when(pullPoint.getMessages(any(), any(), any())).thenThrow(new UnableToGetMessagesFault());

        W3CEndpointReference endpointReference = getEndpointReference("http://pull.point");

        UUID messageId = UUID.randomUUID();
        assertThrows(NasException.class, () -> nasClient.getMessages(endpointReference, 1000L, messageId));
    }

    private W3CEndpointReference getEndpointReference(String address) {
        return new W3CEndpointReferenceBuilder().address(address).build();
    }

    private Header getHeader() {
        Header header = new Header();
        header.setLinking(new Linking());
        return header;
    }


}
