/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidation;

import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.NasLock;
import dk.sundhedsdatastyrelsen.cbs.dao.NasStatus;
import dk.sundhedsdatastyrelsen.cbs.dao.NasSubscription;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.CacheInvalidatingNasNotificationProcessor;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.DefaultRecordMapper;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.LabResultRecordInvalidator;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.LockProvider;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.LockProviderImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.PersistedNasRepository;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.VaccinationMessageRecordMapper;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl.VaccinationRecordInvalidator;
import dk.sundhedsdatastyrelsen.cbs.dgws.DGWSRequestFactory;
import dk.sundhedsdatastyrelsen.cbs.dgws.STSConfig;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientService;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpClientServiceImpl;
import dk.sundhedsdatastyrelsen.cbs.dgws.http.HttpExecutor;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.DGWSHeaderProvider;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.DefaultDGWSHeaderProvider;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClient;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClientImpl;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasConsumer;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasConsumerProperties;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasNotificationProcessor;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasSubscriptionRepository;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.PullPointFactory;
import dk.sundhedsdatastyrelsen.kids.dao.KidsLabResult;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import org.oasis_open.docs.wsn.bw_2.CreatePullPoint;
import org.oasis_open.docs.wsn.bw_2.NotificationProducer;
import org.oasis_open.docs.wsn.bw_2.PullPointFactoryService;
import org.oasis_open.docs.wsn.bw_2.SubscriptionManagerService;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.PropertySource;

import javax.xml.bind.JAXBException;
import javax.xml.ws.BindingProvider;
import java.time.Clock;
import java.util.Collection;

@Configuration
@PropertySource("classpath:application-datacache-invalidator.properties")
public class DataCacheInvalidationConfiguration {


    public static final String VACCINATION_NAS_CONSUMER_BEAN = "vaccinationNasConsumer";
    public static final String LAB_RESULT_NAS_CONSUMER_BEAN = "labResultNasConsumer";

    @Bean
    @Qualifier(VACCINATION_NAS_CONSUMER_BEAN)
    NasConsumer vaccinationNasConsumer(
        NasConsumerProperties vaccinationNasConsumerProperties,
        NasClient nasClient,
        NasSubscriptionRepository repository,
        NasNotificationProcessor vaccinationNotificationProcessor
    ) {
        return new NasConsumer(vaccinationNasConsumerProperties, nasClient, repository, vaccinationNotificationProcessor);
    }

    @Bean
    NasConsumerProperties vaccinationNasConsumerProperties(
        @Value("${vacc.topic}") String topic,
        @Value("${vacc.chunk.size}") long chunkSize,
        @Value("${vacc.chunk.limit}") int chunkMaxLimit
    ) {
        return new NasConsumerProperties(topic, chunkSize, chunkMaxLimit);
    }

    @Bean
    NasNotificationProcessor vaccinationNotificationProcessor(VaccinationRecordInvalidator invalidator, @Value("${vacc.offset.minutes:75}") long offset) {
        return new CacheInvalidatingNasNotificationProcessor(invalidator, new VaccinationMessageRecordMapper(offset));
    }

    @Bean
    VaccinationRecordInvalidator vaccinationRecordInvalidator(Writer<VaccinationRecord<Vaccination, GetVaccinationPassportAsPDFResponseType>> writer) {
        return new VaccinationRecordInvalidator(writer);
    }

    @Bean
    @Qualifier(LAB_RESULT_NAS_CONSUMER_BEAN)
    NasConsumer labResultNasConsumer(
        NasConsumerProperties labResultNasConsumerProperties,
        NasClient nasClient,
        NasSubscriptionRepository repository,
        NasNotificationProcessor labResultNotificationProcessor
    ) {
        return new NasConsumer(labResultNasConsumerProperties, nasClient, repository, labResultNotificationProcessor);
    }

    @Bean
    NasConsumerProperties labResultNasConsumerProperties(
        @Value("${lab.topic}") String topic,
        @Value("${lab.chunk.size}") long chunkSize,
        @Value("${lab.chunk.limit}") int chunkMaxLimit
    ) {
        return new NasConsumerProperties(topic, chunkSize, chunkMaxLimit);
    }

    @Bean
    NasNotificationProcessor labResultNotificationProcessor(LabResultRecordInvalidator invalidator, @Value("${lab.offset.minutes:75}") long offset) {
        return new CacheInvalidatingNasNotificationProcessor(invalidator, new DefaultRecordMapper(offset));
    }

    @Bean
    LabResultRecordInvalidator labResultRecordInvalidator(Writer<LabResultRecord<LabResults, Collection<KidsLabResult>>> writer) {
        return new LabResultRecordInvalidator(writer);
    }

    @Bean
    NasSubscriptionRepository repository(NasClient nasClient,
                                         Reader<NasSubscription, String> subscriptionReader,
                                         Writer<NasSubscription> subscriptionWriter,
                                         Reader<NasStatus, String> statusReader,
                                         WriterWithDelete<NasStatus, String> statusWriter) {
        return new PersistedNasRepository(nasClient, subscriptionReader, subscriptionWriter, statusReader, statusWriter);
    }

    @Bean
    NasClient nasClient(DGWSHeaderProvider dgwsHeaderProvider, NotificationProducer notificationProducer, CreatePullPoint createPullPoint) {
        return new NasClientImpl(dgwsHeaderProvider, notificationProducer, createPullPoint, new PullPointFactory());
    }

    @Bean
    DGWSHeaderProvider dgwsHeaderProvider(DGWSRequestFactory requestFactory) throws JAXBException {
        return new DefaultDGWSHeaderProvider(requestFactory);
    }

    @Bean
    DGWSRequestFactory dgwsRequestFactory(STSConfig stsConfig, HttpClientService httpClientService) {
        return new DGWSRequestFactory(stsConfig, httpClientService);
    }

    @Bean
    HttpClientService httpClientService() {
        return new HttpClientServiceImpl(new HttpExecutor());
    }

    @Bean
    NotificationProducer notificationProducer(@Value("${nas.subscriptionmanager.endpoint}") String endpoint) {
        NotificationProducer notificationProducer = new SubscriptionManagerService().getSoap();

        BindingProvider bindingProvider = (BindingProvider) notificationProducer;

        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpoint);

        return notificationProducer;
    }

    @Bean
    CreatePullPoint createPullPoint(@Value("${nas.pullpointfactory.endpoint}") String endpoint) {
        CreatePullPoint createPullPoint = new PullPointFactoryService().getSoap();

        BindingProvider bindingProvider = (BindingProvider) createPullPoint;

        bindingProvider.getRequestContext().put(BindingProvider.ENDPOINT_ADDRESS_PROPERTY, endpoint);

        return createPullPoint;
    }

    @Bean
    LockProvider lockProvider(Reader<NasLock, String> reader, WriterWithDelete<NasLock, String> writer, @Value("${nas.lock.at-least.minutes:3}") long lockForAtLeastMinutes){
        return new LockProviderImpl(reader, writer, lockForAtLeastMinutes, Clock.systemDefaultZone());
    }

    @Bean
    public STSConfig stsConfig(
        @Value("${nas.idcard.sts.endpoint}") String stsEndpoint,
        @Value("${nas.idcard.name}") String organizationName,
        @Value("${nas.idcard.cvr}") String organizationCode,
        @Value("${nas.idcard.keystore.path}") String keystorePath,
        @Value("${nas.idcard.keystore.password}") String keystorePassword,
        @Value("${nas.idcard.systemName}") String systemName) {
        return new STSConfig(stsEndpoint, organizationName, organizationCode, keystorePath, keystorePassword, systemName);
    }

}
