/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl;

import dk.sundhedsdatastyrelsen.cbs.nas.facade.EndpointReferences;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClient;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasSubscriptionRepository;
import org.apache.log4j.Logger;

import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

public class InMemoryNasSubscriptionRepository implements NasSubscriptionRepository {

    private static final Logger log = Logger.getLogger(InMemoryNasSubscriptionRepository.class);

    private final NasClient nasClient;

    private final Map<String, String> pullPointMap = new HashMap<>();
    private final Map<String, UUID> messageIdMap = new HashMap<>();


    public InMemoryNasSubscriptionRepository(NasClient nasClient) {
        this.nasClient = nasClient;
    }

    @Override
    public W3CEndpointReference getPullPoint(String topic) {
        if (pullPointMap.containsKey(topic)) {
            String address = pullPointMap.get(topic);

            log.info(String.format("Found saved PullPoint (%s)", address));

            return EndpointReferences.newEndpointReference(address);
        } else {
            log.info("No PullPoint found");

            W3CEndpointReference pullPoint = nasClient.createPullPoint();
            nasClient.subscribe(pullPoint, topic);
            String address = EndpointReferences.extractAddress(pullPoint);

            log.info(String.format("Created new PullPoint (%s)", address));

            pullPointMap.put(topic, address);
            return pullPoint;
        }
    }

    @Override
    public UUID getNextMessageId(String topic) {
        if (messageIdMap.containsKey(topic)) {
            UUID messageId = messageIdMap.get(topic);
            log.info(String.format("Found saved MessageId (%s)", messageId));
            return messageId;
        } else {
            UUID messageId = UUID.randomUUID();
            messageIdMap.put(topic, messageId);
            log.info(String.format("Created new MessageId (%s)", messageId));
            return messageId;
        }
    }

    @Override
    public void commit(String topic) {
        log.info(String.format("Commit on topic '%s'", topic));

        messageIdMap.remove(topic);
    }
}
