/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.NasStatus;
import dk.sundhedsdatastyrelsen.cbs.dao.NasSubscription;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasStatusImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasSubscriptionImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.EndpointReferences;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasClient;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasSubscriptionRepository;
import org.apache.log4j.Logger;
import org.springframework.dao.EmptyResultDataAccessException;

import javax.xml.ws.wsaddressing.W3CEndpointReference;
import java.util.UUID;

public class PersistedNasRepository implements NasSubscriptionRepository {

    private static final Logger log = Logger.getLogger(PersistedNasRepository.class);


    private final NasClient nasClient;

    private final Reader<NasSubscription, String> subscriptionReader;
    private final Writer<NasSubscription> subscriptionWriter;

    private final Reader<NasStatus, String> statusReader;
    private final WriterWithDelete<NasStatus, String> statusWriter;

    public PersistedNasRepository(NasClient nasClient,
                                  Reader<NasSubscription, String> subscriptionReader,
                                  Writer<NasSubscription> subscriptionWriter,
                                  Reader<NasStatus, String> statusReader,
                                  WriterWithDelete<NasStatus, String> statusWriter) {
        this.nasClient = nasClient;
        this.subscriptionReader = subscriptionReader;
        this.subscriptionWriter = subscriptionWriter;
        this.statusReader = statusReader;
        this.statusWriter = statusWriter;
    }


    @Override
    public W3CEndpointReference getPullPoint(String topic) {
        NasSubscription subscription = getSubscription(topic);

        if (subscription == null) {

            W3CEndpointReference pullPoint = nasClient.createPullPoint();
            W3CEndpointReference subscribe = nasClient.subscribe(pullPoint, topic);

            subscription = new NasSubscriptionImpl()
                .topic(topic)
                .pullPoint(EndpointReferences.extractAddress(pullPoint))
                .subscription(EndpointReferences.extractAddress(subscribe));

            subscriptionWriter.write(subscription);

            log.debug(String.format("Created new pull point '%s' for topic '%s'", subscription.pullPoint(), topic));
        } else {
            log.debug(String.format("Found persisted pull point '%s' for topic '%s'", subscription.pullPoint(), topic));
        }

        return EndpointReferences.newEndpointReference(subscription.pullPoint());
    }

    private NasSubscription getSubscription(String topic) {
        try {
            return subscriptionReader.search(topic);
        } catch (EmptyResultDataAccessException e) {
            return null;
        }
    }

    @Override
    public UUID getNextMessageId(String topic) {
        NasStatus status = getNasStatus(topic);

        if (status == null) {
            status = new NasStatusImpl().topic(topic).messageId(UUID.randomUUID());
            statusWriter.write(status);
            log.debug(String.format("Created new messageId '%s' for topic '%s'", status.messageId(), topic));
        } else {
            log.debug(String.format("Found persisted messageId '%s' for topic '%s'", status.messageId(), topic));
        }

        return status.messageId();
    }

    private NasStatus getNasStatus(String topic) {
        try {
            return statusReader.search(topic);
        } catch (EmptyResultDataAccessException e) {
            return null;
        }
    }

    @Override
    public void commit(String topic) {
        statusWriter.delete(topic);
    }
}
