/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.impl;

import dk.nsi.advis.v10.NotifyContent;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.CacheInvalidationRecord;
import dk.sundhedsdatastyrelsen.cbs.datacache.invalidation.RecordInvalidator;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.Constants;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.oasis_open.docs.wsn.b_2.NotificationMessageHolderType;
import org.oasis_open.docs.wsn.b_2.TopicExpressionType;

import java.time.LocalDateTime;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoInteractions;
import static org.mockito.Mockito.when;

@CbsTest
@SuppressWarnings("java:S5979") /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
class CacheInvalidatingNasNotificationProcessorTest {

    private final org.oasis_open.docs.wsn.b_2.ObjectFactory b2ObjectFactory = new org.oasis_open.docs.wsn.b_2.ObjectFactory();
    private final dk.nsi.advis.v10.ObjectFactory nsiObjectFactory = new dk.nsi.advis.v10.ObjectFactory();

    @Mock
    private RecordInvalidator recordInvalidator;

    @Captor
    private ArgumentCaptor<CacheInvalidationRecord> recordArgumentCaptor;


    private CacheInvalidatingNasNotificationProcessor processor;

    @BeforeEach
    void setUp() {
        processor = new CacheInvalidatingNasNotificationProcessor(recordInvalidator, new DefaultRecordMapper(10L));
    }

    @Test
    void should_invalidate_records() {
        when(recordInvalidator.invalidate(any())).thenReturn(true);

        String cpr = "0506504003";
        processor.process(Collections.singletonList(getMessage(cpr)));

        verify(recordInvalidator).invalidate(recordArgumentCaptor.capture());

        CacheInvalidationRecord record = recordArgumentCaptor.getValue();

        LocalDateTime now = LocalDateTime.now();

        assertEquals(cpr, record.getCpr());
        assertTrue(now.isBefore(record.getInvalidationTime()));
        assertTrue(now.plusMinutes(11L).isAfter(record.getInvalidationTime()));
    }

    @Test
    void should_throw_if_record_is_not_invalidated() {
        when(recordInvalidator.invalidate(any())).thenReturn(false);

        List<NotificationMessageHolderType> messages = Collections.singletonList(getMessage("0506504003"));
        assertThrows(DataCacheInvalidationException.class, () -> processor.process(messages));
    }

    @Test
    void should_throw_if_record_invalidation_fails() {
        when(recordInvalidator.invalidate(any())).thenThrow(new IllegalStateException());

        List<NotificationMessageHolderType> messages = Collections.singletonList(getMessage("0506504003"));
        assertThrows(IllegalStateException.class, () -> processor.process(messages));
    }

    @Test
    void should_skip_notifications_with_no_content() {
        processor.process(Collections.singletonList(emptyMessage()));
        verifyNoInteractions(recordInvalidator);
    }

    @Test
    void should_skip_notifications_with_unknown_content() {
        processor.process(Collections.singletonList(unknownMessage()));
        verifyNoInteractions(recordInvalidator);
    }

    @Test
    void should_process_until_complete() {
        when(recordInvalidator.invalidate(any())).thenReturn(true);

        List<NotificationMessageHolderType> messages = Stream.generate(() -> getMessage("0506504003")).limit(50).collect(Collectors.toList());

        processor.process(messages);

        verify(recordInvalidator, times(50)).invalidate(any());
    }

    @Test
    void should_process_until_first_record_cant_be_invalidated() {
        when(recordInvalidator.invalidate(any())).thenReturn(true, true, true, false);

        List<NotificationMessageHolderType> messages = Stream.generate(() -> getMessage("0506504003")).limit(10).collect(Collectors.toList());

        assertThrows(DataCacheInvalidationException.class, () -> processor.process(messages));

        verify(recordInvalidator, times(4)).invalidate(any());
    }

    @Test
    void should_process_until_first_record_invalidation_fails() {
        when(recordInvalidator.invalidate(any())).thenReturn(true, true, true).thenThrow(new IllegalStateException());

        List<NotificationMessageHolderType> messages = Stream.generate(() -> getMessage("0506504003")).limit(10).collect(Collectors.toList());

        assertThrows(IllegalStateException.class, () -> processor.process(messages));

        verify(recordInvalidator, times(4)).invalidate(any());
    }

    private NotificationMessageHolderType unknownMessage() {
        NotificationMessageHolderType.Message message = b2ObjectFactory.createNotificationMessageHolderTypeMessage();
        message.setAny("This is an unknown message type");

        TopicExpressionType topicExpressionType = b2ObjectFactory.createTopicExpressionType();
        topicExpressionType.setDialect(Constants.TOPIC_EXPRESSION_SIMPLE);
        topicExpressionType.getContent().add("myTopic");

        NotificationMessageHolderType notificationMessageHolderType = b2ObjectFactory.createNotificationMessageHolderType();
        notificationMessageHolderType.setMessage(message);
        notificationMessageHolderType.setTopic(topicExpressionType);
        return notificationMessageHolderType;
    }

    private NotificationMessageHolderType emptyMessage() {
        NotificationMessageHolderType.Message message = b2ObjectFactory.createNotificationMessageHolderTypeMessage();
        message.setAny(null);

        TopicExpressionType topicExpressionType = b2ObjectFactory.createTopicExpressionType();
        topicExpressionType.setDialect(Constants.TOPIC_EXPRESSION_SIMPLE);
        topicExpressionType.getContent().add("myTopic");

        NotificationMessageHolderType notificationMessageHolderType = b2ObjectFactory.createNotificationMessageHolderType();
        notificationMessageHolderType.setMessage(message);
        notificationMessageHolderType.setTopic(topicExpressionType);
        return notificationMessageHolderType;
    }

    private NotificationMessageHolderType getMessage(String cpr) {
        NotifyContent notifyContent = nsiObjectFactory.createNotifyContent();
        notifyContent.setAny("payload");
        notifyContent.setId(cpr);
        notifyContent.setIdType(Constants.ID_TYPE_CPR);
        notifyContent.setIsSystemNotification(false);

        NotificationMessageHolderType.Message message = b2ObjectFactory.createNotificationMessageHolderTypeMessage();
        message.setAny(notifyContent);

        TopicExpressionType topicExpressionType = b2ObjectFactory.createTopicExpressionType();
        topicExpressionType.setDialect(Constants.TOPIC_EXPRESSION_SIMPLE);
        topicExpressionType.getContent().add("myTopic");

        NotificationMessageHolderType notificationMessageHolderType = b2ObjectFactory.createNotificationMessageHolderType();
        notificationMessageHolderType.setMessage(message);
        notificationMessageHolderType.setTopic(topicExpressionType);
        return notificationMessageHolderType;
    }
}
