/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc;

import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mappers;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.ResultSets;
import org.springframework.jdbc.core.support.SqlLobValue;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.io.Reader;
import java.sql.Types;
import java.util.function.Function;
import java.util.function.Supplier;

import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordImpl.CPR_COLUMN_LABEL;
import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordImpl.CREATION_TIME_COLUMN_LABEL;
import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordImpl.DATA_COLUMN_LABEL;
import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordImpl.INVALIDATION_TIME_COLUMN_LABEL;
import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordImpl.SOURCE_COLUMN_LABEL;
import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordImpl.TRACE_ID_COLUMN_LABEL;
import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mappers.deserializer;
import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mappers.serializer;
import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaDataBuilder.tableMetaData;

public final class RecordMappers {

    private RecordMappers() {
    }

    public static <T, U, V extends RecordImpl<T, U, V>> Mappers.KeyMapperBuilder<V, String> mapper(
        @Nullable final String insertCatalogName,
        @Nullable final String insertSchemaName,
        @NonNull final String insertTableName,
        @Nullable final String selectCatalogName,
        @Nullable final String selectSchemaName,
        @NonNull final String selectTableName,
        @NonNull final Supplier<V> constructor,
        @NonNull final Function<Reader, T> dataDeserializer,
        @NonNull final Function<T, String> dataSerializer,
        @NonNull final Function<Reader, U> sourceDeserializer,
        @NonNull final Function<U, String> sourceSerializer) {
        return Mappers.mapper(
            String.class,
            tableMetaData(insertTableName).schemaName(insertSchemaName).catalogName(insertCatalogName).build(),
            tableMetaData(selectTableName).schemaName(selectSchemaName).catalogName(selectCatalogName).build(),
            constructor)
            .addKeyStringMapping(CPR_COLUMN_LABEL, RecordImpl::cpr, RecordImpl::cpr, Function.identity())
            .addStringMapping(TRACE_ID_COLUMN_LABEL, RecordImpl::traceId, RecordImpl::traceId)
            .addTimestampToLocalDateTimeMapping(CREATION_TIME_COLUMN_LABEL, RecordImpl::creationTime, RecordImpl::creationTime)
            .addTimestampToLocalDateTimeMapping(INVALIDATION_TIME_COLUMN_LABEL, RecordImpl::invalidationTime, RecordImpl::invalidationTime)
            .addMapping(
                DATA_COLUMN_LABEL,
                deserializer(ResultSets::getCharacterStream).map(dataDeserializer).set(V::setData),
                serializer(V::data).map(dataSerializer).get(SqlLobValue::new, Types.CLOB))
            .addMapping(
                SOURCE_COLUMN_LABEL,
                deserializer(ResultSets::getCharacterStream).map(sourceDeserializer).set(V::setSource),
                serializer(V::source).map(sourceSerializer).get(SqlLobValue::new, Types.CLOB));
    }

    public static <V extends NasSubscriptionImpl> Mappers.KeyMapperBuilder<V, String> nasSubscriptionMapper(
        @Nullable final String insertCatalogName,
        @Nullable final String insertSchemaName,
        @NonNull final String insertTableName,
        @Nullable final String selectCatalogName,
        @Nullable final String selectSchemaName,
        @NonNull final String selectTableName,
        @NonNull final Supplier<V> constructor) {
        return Mappers.mapper(
            String.class,
            tableMetaData(insertTableName).schemaName(insertSchemaName).catalogName(insertCatalogName).build(),
            tableMetaData(selectTableName).schemaName(selectSchemaName).catalogName(selectCatalogName).build(),
            constructor)
            .addKeyStringMapping(NasSubscriptionImpl.TOPIC_COLUMN_LABEL, NasSubscriptionImpl::topic, NasSubscriptionImpl::topic, Function.identity())
            .addStringMapping(NasSubscriptionImpl.PULL_POINT_COLUMN_LABEL, NasSubscriptionImpl::pullPoint, NasSubscriptionImpl::pullPoint)
            .addStringMapping(NasSubscriptionImpl.SUBSCRIPTION_COLUMN_LABEL, NasSubscriptionImpl::subscription, NasSubscriptionImpl::subscription);
    }

    public static <V extends NasStatusImpl> Mappers.KeyMapperBuilder<V, String> nasStatusMapper(
        @Nullable final String insertCatalogName,
        @Nullable final String insertSchemaName,
        @NonNull final String insertTableName,
        @Nullable final String selectCatalogName,
        @Nullable final String selectSchemaName,
        @NonNull final String selectTableName,
        @NonNull final Supplier<V> constructor) {
        return Mappers.mapper(
            String.class,
            tableMetaData(insertTableName).schemaName(insertSchemaName).catalogName(insertCatalogName).build(),
            tableMetaData(selectTableName).schemaName(selectSchemaName).catalogName(selectCatalogName).build(),
            constructor)
            .addKeyStringMapping(NasStatusImpl.TOPIC_COLUMN_LABEL, NasStatusImpl::topic, NasStatusImpl::topic, Function.identity())
            .addBytesMapping(NasStatusImpl.MESSAGE_ID_COLUMN_LABEL, NasStatusImpl::messageId, NasStatusImpl::messageId);
    }

    public static <V extends NasLockImpl> Mappers.KeyMapperBuilder<V, String> nasLockMapper(
        @Nullable final String insertCatalogName,
        @Nullable final String insertSchemaName,
        @NonNull final String insertTableName,
        @Nullable final String selectCatalogName,
        @Nullable final String selectSchemaName,
        @NonNull final String selectTableName,
        @NonNull final Supplier<V> constructor) {
        return Mappers.mapper(
            String.class,
            tableMetaData(insertTableName).schemaName(insertSchemaName).catalogName(insertCatalogName).build(),
            tableMetaData(selectTableName).schemaName(selectSchemaName).catalogName(selectCatalogName).build(),
            constructor)
            .addKeyStringMapping(NasLockImpl.TOPIC_COLUMN_LABEL, NasLockImpl::topic, NasLockImpl::topic, Function.identity())
            .addTimestampToLocalDateTimeMapping(NasLockImpl.LOCKED_UNTIL_COLUMN_LABEL, NasLockImpl::lockedUntil, NasLockImpl::lockedUntil);
    }

}
