/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache;

import com.fasterxml.jackson.databind.ObjectMapper;
import dk.sundhedsdatastyrelsen.cbs.dao.LabResultRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.NasLock;
import dk.sundhedsdatastyrelsen.cbs.dao.NasStatus;
import dk.sundhedsdatastyrelsen.cbs.dao.NasSubscription;
import dk.sundhedsdatastyrelsen.cbs.dao.Record;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.LabResultRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasLockImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasStatusImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.NasSubscriptionImpl;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.kids.dao.KidsLabResult;
import dk.sundhedsdatastyrelsen.kids.dao.KidsLabResultImpl;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.RecordMappers.mapper;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Readers.reader;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Readers.recordReader;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Writers.writer;
import static dk.sundhedsdatastyrelsen.cbs.datacache.impl.Writers.writerWithDelete;

@Configuration
public class DataCacheImplConfiguration {

    @Bean
    @NonNull
    public Reader<Vaccination, String> vaccinationReader(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<VaccinationRecordImpl<Vaccination, GetVaccinationPassportAsPDFResponseType>, String> mapper) {
        return recordReader(dataSource, mapper);
    }

    @Bean
    @NonNull
    public Writer<VaccinationRecord<Vaccination, GetVaccinationPassportAsPDFResponseType>> vaccinationWriter(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<VaccinationRecordImpl<Vaccination, GetVaccinationPassportAsPDFResponseType>, String> mapper) {
        return writer(dataSource, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<VaccinationRecordImpl<Vaccination, GetVaccinationPassportAsPDFResponseType>, String> vaccinationMapper(
        @NonNull final Supplier<VaccinationRecordImpl<Vaccination, GetVaccinationPassportAsPDFResponseType>> constructor,
        @NonNull final ObjectMapper mapper) {
        return mapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            VaccinationRecord.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            VaccinationRecord.SELECT_TABLE_NAME,
            constructor,
            deserializer(mapper, Vaccination.class),
            serializer(mapper),
            deserializer(mapper, GetVaccinationPassportAsPDFResponseType.class),
            serializer(mapper))
            .build();
    }

    @Bean
    @NonNull
    public Reader<LabResults, String> labResultReader(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<LabResultRecordImpl<LabResults, Collection<KidsLabResult>>, String> mapper) {
        return recordReader(dataSource, mapper);
    }


    @Bean
    @NonNull
    public Writer<LabResultRecord<LabResults, Collection<KidsLabResult>>> labResultWriter(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<LabResultRecordImpl<LabResults, Collection<KidsLabResult>>, String> mapper) {
        return writer(dataSource, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<LabResultRecordImpl<LabResults, Collection<KidsLabResult>>, String> labResultMapper(
        @NonNull final Supplier<LabResultRecordImpl<LabResults, Collection<KidsLabResult>>> constructor,
        @NonNull final ObjectMapper mapper) {
        return mapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            LabResultRecord.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            LabResultRecord.SELECT_TABLE_NAME,
            constructor,
            deserializer(mapper, LabResults.class),
            serializer(mapper),
            collectionDeserializer(mapper),
            serializer(mapper))
            .build();
    }

    @Bean
    @NonNull
    public Supplier<VaccinationRecordImpl<Vaccination, GetVaccinationPassportAsPDFResponseType>> vaccinationConstructor() {
        return VaccinationRecordImpl::new;
    }

    @Bean
    @NonNull
    public Supplier<LabResultRecordImpl<LabResults, Collection<KidsLabResult>>> labResultConstructor() {
        return LabResultRecordImpl::new;
    }


    @Bean
    @NonNull
    public KeyMapper<NasLockImpl, String> nasLockMapper() {
        return RecordMappers.nasLockMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasLock.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasLock.SELECT_TABLE_NAME,
            NasLockImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<NasLock, String> nasLockReader(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<NasLockImpl, String> mapper) {
        return reader(dataSource, mapper);
    }

    @Bean
    @NonNull
    public WriterWithDelete<NasLock, String> nasLockWriter(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<NasLockImpl, String> mapper) {
        return writerWithDelete(dataSource, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<NasStatusImpl, String> nasStatusMapper() {
        return RecordMappers.nasStatusMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasStatus.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasStatus.SELECT_TABLE_NAME,
            NasStatusImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<NasStatus, String> nasStatusReader(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<NasStatusImpl, String> mapper) {
        return reader(dataSource, mapper);
    }

    @Bean
    @NonNull
    public WriterWithDelete<NasStatus, String> nasStatusWriter(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<NasStatusImpl, String> mapper) {
        return writerWithDelete(dataSource, mapper);
    }

    @Bean
    @NonNull
    public KeyMapper<NasSubscriptionImpl, String> nasSubscriptionMapper() {
        return RecordMappers.nasSubscriptionMapper(
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasSubscription.INSERT_TABLE_NAME,
            Record.CATALOG_NAME,
            Record.SCHEMA_NAME,
            NasSubscription.SELECT_TABLE_NAME,
            NasSubscriptionImpl::new)
            .build();
    }

    @Bean
    @NonNull
    public Reader<NasSubscription, String> nasSubscriptionReader(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<NasSubscriptionImpl, String> mapper) {
        return reader(dataSource, mapper);
    }

    @Bean
    @NonNull
    public Writer<NasSubscription> nasSubscriptionWriter(
        @NonNull final DataSource dataSource,
        @NonNull final KeyMapper<NasSubscriptionImpl, String> mapper) {
        return writer(dataSource, mapper);
    }

    private static <T> Function<java.io.Reader, T> deserializer(@NonNull final ObjectMapper mapper, @NonNull final Class<T> clazz) {
        return reader -> {
            try {
                return reader == null ? null : mapper.readValue(reader, clazz);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }

    private static Function<java.io.Reader, Collection<KidsLabResult>> collectionDeserializer(@NonNull final ObjectMapper mapper) {
        return reader -> {
            try {
                return reader == null ? null : Arrays.stream(mapper.readValue(reader, KidsLabResultImpl[].class)).collect(Collectors.toList());
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }


    private static <T> Function<T, String> serializer(@NonNull final ObjectMapper mapper) {
        return object -> {
            try {
                return object == null ? null : mapper.writeValueAsString(object);
            } catch (IOException e) {
                throw new RuntimeException(e);
            }
        };
    }

}
