/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.impl;

import dk.sundhedsdatastyrelsen.cbs.dao.Record;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.SearchMapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import javax.sql.DataSource;
import java.util.Collection;

public final class Readers {

    private Readers() {
    }

    public static <T, K> Reader<T, K> reader(@NonNull final DataSource dataSource, @NonNull final KeyMapper<? extends T, K> mapper) {
        return new KeyReaderImpl<>(dataSource, mapper);
    }

    public static <T, K> Reader<Collection<T>, K> reader(@NonNull final DataSource dataSource, @NonNull final SearchMapper<? extends T, K> mapper) {
        return new SearchReaderImpl<>(dataSource, mapper);
    }

    public static <T, K> Reader<T, K> recordReader(@NonNull final DataSource dataSource, @NonNull final KeyMapper<? extends Record<T, ?>, K> mapper) {
        return new RecordReaderImpl<>(dataSource, mapper);
    }

    private static final class SearchReaderImpl<T, S> implements Reader<Collection<T>, S> {
        @NonNull private final SearchMapper<T, S> mapper;
        @NonNull private final JdbcTemplate template;
        @NonNull private final String query;

        @SuppressWarnings("unchecked")
        private SearchReaderImpl(
            @NonNull final DataSource dataSource,
            @NonNull final SearchMapper<? extends T, S> mapper) {
            template = new JdbcTemplate(dataSource);
            query = "select * from " + mapper.selectTableMetaData().tableName() + " where " + mapper.searchColumnLabel() + " = ?";
            this.mapper = (SearchMapper<T, S>) mapper;
        }

        @Override
        @NonNull
        public Collection<T> search(@NonNull S value) {
            return template.query(query, mapper, mapper.serializeSearchValue(value));
        }

    }

    private static final class KeyReaderImpl<T, K> implements Reader<T, K> {
        @NonNull private final KeyMapper<? extends T, K> mapper;
        @NonNull private final JdbcTemplate template;
        @NonNull private final String query;

        private KeyReaderImpl(
            @NonNull final DataSource dataSource,
            @NonNull final KeyMapper<? extends T, K> mapper) {
            template = new JdbcTemplate(dataSource);
            query = "select * from " + mapper.selectTableMetaData().tableName() + " where " + mapper.keyColumnLabel() + " = ?";
            this.mapper = mapper;
        }

        @Override
        @Nullable
        public T search(@NonNull K key) {
            return template.queryForObject(query, mapper, mapper.serializeKey(key));
        }
    }

    private static final class RecordReaderImpl<T, K> implements Reader<T, K> {

        @NonNull
        private final JdbcTemplate template;

        @NonNull
        private final KeyMapper<? extends Record<T, ?>, K> mapper;

        @NonNull
        private final String query;

        private RecordReaderImpl(@NonNull final DataSource dataSource, @NonNull final KeyMapper<? extends Record<T, ?>, K> mapper) {
            template = new JdbcTemplate(dataSource);
            query = "select * from " + mapper.selectTableMetaData().tableName() + " where " + mapper.keyColumnLabel() + " = ?";
            this.mapper = mapper;
        }

        @Override
        public T search(@NonNull final K key) {
            return template.query(query, mapper, mapper.serializeKey(key)).stream().findFirst()
                .map(Record::data)
                .orElse(null);
        }
    }
}
