/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.impl;

import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datacache.WriterWithDelete;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.KeyMapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mapper;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.simple.SimpleJdbcInsert;
import org.springframework.lang.NonNull;

import javax.sql.DataSource;

public final class Writers {
    private Writers() {
    }

    @NonNull
    public static <T> Writer<T> writer(@NonNull final DataSource dataSource, @NonNull final Mapper<? extends T> mapper) {
        return new WriterImpl<>(dataSource, mapper);
    }

    @NonNull
    public static <T, K> WriterWithDelete<T, K> writerWithDelete(@NonNull final DataSource dataSource, @NonNull final KeyMapper<? extends T, K> mapper) {
        return new WriterWithDeleteImpl<>(dataSource, mapper);
    }

    private static class WriterImpl<T> implements Writer<T> {

        protected static final int EXPECTED_ROWS_AFFECTED = 1;
        @NonNull
        private final DataSource dataSource;
        @NonNull
        private final Mapper<T> mapper;

        @SuppressWarnings("unchecked")
        private WriterImpl(@NonNull final DataSource dataSource, @NonNull final Mapper<? extends T> mapper) {
            this.dataSource = dataSource;
            this.mapper = (Mapper<T>) mapper;
        }

        @Override
        @SuppressWarnings("ConstantConditions")
        public boolean write(@NonNull T obj) {
            return EXPECTED_ROWS_AFFECTED == new SimpleJdbcInsert(dataSource)
                .withCatalogName(mapper.insertTableMetaData().catalogName())
                .withSchemaName(mapper.insertTableMetaData().schemaName())
                .withTableName(mapper.insertTableMetaData().tableName())
                .execute(mapper.parameterSource(obj));
        }
    }

    private static final class WriterWithDeleteImpl<T, K> extends WriterImpl<T> implements WriterWithDelete<T, K> {

        @NonNull
        private final JdbcTemplate jdbcTemplate;
        @NonNull
        private final String sql;

        @NonNull
        private final KeyMapper<? extends T, K> keyMapper;

        private WriterWithDeleteImpl(@NonNull final DataSource dataSource, @NonNull final KeyMapper<? extends T, K> mapper) {
            super(dataSource, mapper);
            jdbcTemplate = new JdbcTemplate(dataSource);
            sql = "delete from " + mapper.selectTableMetaData().tableName() + " where " + mapper.keyColumnLabel() + " = ?";
            this.keyMapper = mapper;
        }

        @Override
        public boolean delete(K key) {
            return EXPECTED_ROWS_AFFECTED == jdbcTemplate.update(sql, keyMapper.serializeKey(key));
        }
    }
}
