/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.impl;

import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mapper;
import dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaData;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Types;
import java.util.concurrent.atomic.AtomicBoolean;

import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaDataBuilder.tableMetaData;
import static java.lang.String.format;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

@ExtendWith(MockitoExtension.class)
class WritersTest {

    private static final String SCHEMA_NAME = "SCHEMA";
    private static final String TABLE_NAME = "TABLE";
    private static final String COLUMN_NAME = "COLUMN";
    private static final String EXPECTED_SQL =
        format("INSERT INTO %s.%s (%s) VALUES(?)", SCHEMA_NAME, TABLE_NAME, COLUMN_NAME);
    private final Writer<String> writer;

    WritersTest(
        @Mock final DataSource dataSource,
        @Mock final Mapper<String> mapper,
        @Mock final Connection connection,
        @Mock final DatabaseMetaData metaData,
        @Mock final ResultSet tableResultSet,
        @Mock final ResultSet columnResultSet,
        @Mock final PreparedStatement preparedStatement) {
        writer = Writers.writer(dataSource, mapper);
        final AtomicBoolean nextTable = new AtomicBoolean(true);
        final AtomicBoolean nextColumn = new AtomicBoolean(true);
        final TableMetaData tableMetaData = tableMetaData(TABLE_NAME).schemaName(SCHEMA_NAME).build();
        try {
            when(mapper.insertTableMetaData()).thenReturn(tableMetaData);
            when(dataSource.getConnection()).thenReturn(connection);
            when(connection.getMetaData()).thenReturn(metaData);
            when(metaData.getTables(any(), any(), any(), any())).thenReturn(tableResultSet);
            when(tableResultSet.next()).thenAnswer(i -> nextTable.getAndSet(false));
            when(tableResultSet.getString("TABLE_SCHEM")).thenReturn(SCHEMA_NAME);
            when(tableResultSet.getString("TABLE_NAME")).thenReturn(TABLE_NAME);
            when(metaData.getColumns(any(), any(), any(), any())).thenReturn(columnResultSet);
            when(columnResultSet.next()).thenAnswer(i -> nextColumn.getAndSet(false));
            when(columnResultSet.getString("COLUMN_NAME")).thenReturn(COLUMN_NAME);
            when(columnResultSet.getInt("DATA_TYPE")).thenReturn(Types.VARCHAR);
            when(columnResultSet.getBoolean("NULLABLE")).thenReturn(false);
            when(mapper.parameterSource(any())).thenReturn(new MapSqlParameterSource(COLUMN_NAME, "VALUE"));
            when(connection.prepareStatement(EXPECTED_SQL)).thenReturn(preparedStatement);
            when(preparedStatement.executeUpdate()).thenReturn(1);
        } catch (Exception e) {
            throw new ExceptionInInitializerError(e);
        }
    }

    @Test
    void write() {
        final String record = "Record";
        assertTrue(writer.write(record));
    }
}
