/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datacache.liquibase;

import liquibase.Contexts;
import liquibase.LabelExpression;
import liquibase.Liquibase;
import liquibase.database.DatabaseFactory;
import liquibase.database.jvm.JdbcConnection;
import liquibase.exception.LiquibaseException;
import liquibase.resource.ClassLoaderResourceAccessor;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.jdbc.datasource.DriverManagerDataSource;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

@Service
public class AsyncLiquibaseService {

    @Value("${liquibase.changelog.file}")
    private String changeLogFile;

    @Value("${liquibase.connection.polling.delay}")
    private long pollingDelay;

    @Value("${liquibase.connection.polling.attempts}")
    private int pollingAttempts;

    private final DataSource dataSource;
    private static final Logger log = Logger.getLogger(AsyncLiquibaseService.class);

    public AsyncLiquibaseService(
        @Value("${spring.datasource.driver}") String driverClassName,
        @Value("${spring.datasource.url}") String url,
        @Value("${liquibase.datasource.username}") String username,
        @Value("${liquibase.datasource.password}") String password
    ) {
        DriverManagerDataSource ds = new DriverManagerDataSource();
        ds.setDriverClassName(driverClassName);
        ds.setUrl(url);
        ds.setUsername(username);
        ds.setPassword(password);
        this.dataSource = ds;

    }

    /**
     * Configures Liquibase so that it can be triggered at runtime after the database is up and running.
     * The DATABASECHANGELOG entry will be tagged with the project version for convenient rollback.
     * <p>
     * The bean polls the database until a connection is established.  This is to ensure the database is up and running
     * before Liquibase attempts to update
     */
    @Async
    public void update() {
        boolean isDatabaseReady = false;
        int attemptsRemaining = pollingAttempts;

        while (!isDatabaseReady && attemptsRemaining > 0) {
            try (Connection connection = dataSource.getConnection()) {
                isDatabaseReady = isDatabaseReady(connection);
                if (isDatabaseReady) {
                    log.info("Updating cbs-datacache database from: " + changeLogFile + "'");
                    try (Liquibase liquibase = new Liquibase(
                        changeLogFile,
                        new ClassLoaderResourceAccessor(),
                        DatabaseFactory.getInstance().findCorrectDatabaseImplementation(new JdbcConnection(connection))
                    )) {
                        liquibase.update(new Contexts(), new LabelExpression());
                    }
                }
            } catch (LiquibaseException liquibaseException) {
                log.warn("LiquibaseException occurred. Change sets may not have been applied correctly: " + liquibaseException);
                attemptsRemaining = 0;
            } catch (SQLException ignored) {
                //try again later
                try {
                    Thread.sleep(pollingDelay);
                } catch (InterruptedException interruptedException) {
                    attemptsRemaining = 0;
                    Thread.currentThread().interrupt();
                }
            } finally {
                attemptsRemaining--;
            }
        }
        if (!isDatabaseReady) {
            log.warn("Database connection failed after " + pollingAttempts + " attempts.  " + "Liquibase change sets may not have been applied.");
        }
    }

    /**
     * Attempt to send a query to the database
     *
     * @param connection to database
     * @return true if select query returns something
     */
    private boolean isDatabaseReady(Connection connection) {
        ResultSet result = null;

        try {
            if (connection == null || connection.isClosed()) {
                return false;
            }

            try (Statement statement = connection.createStatement()) {
                result = statement.executeQuery("select 1");
            }

        } catch (SQLException e) {
            log.debug("DB not ready: " + e.getMessage());
        }

        return result != null;
    }
}
