/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher;

import com.google.common.base.Strings;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.config.DataFetcherConfigProperties;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.model.DataFetcherJobExecutionResult;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.kafka.pooling.CbsKafkaConsumerPool;
import dk.sundhedsdatastyrelsen.cbs.monitoring.IsAliveServlet;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasPublisher;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatus;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatusNotification;
import org.apache.commons.lang3.NotImplementedException;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;

import java.io.Closeable;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Entry point for interacting with the datafetcher inside a Spring ApplicationContext
 */
@Service
@Lazy
public class DataFetcher implements Closeable {
    private static final Logger log = Logger.getLogger(DataFetcher.class);
    private static final String CLOSING_SERVER_JOB_FAILURE_MSG = "INTERNAL_CBS_ERROR: SERVER_SHUTDOWN";
    public static final String VACCINATION_ETL_FAILURE_MSG = "INTERNAL_CBS_ERROR: Vaccination ETL failed";
    public static final String LAB_RESULT_ETL_FAILURE_MSG = "INTERNAL_CBS_ERROR: LabResult ETL failed";
    private final CbsKafkaConsumerPool kafkaConsumerPool;
    private final Function<String, Boolean> vaccinationEtl;
    private final Function<String, Boolean> labResEtl;
    private final int processingEndTimeoutMillis;
    private final int processingBeginTimeoutMillis;
    private final long shutdownTimeoutMillis;
    private final Semaphore processingPermits;
    private Instant shutdownTime = null;
    private final ExecutorService executorService;
    private final NasPublisher nasPublisher;

    public DataFetcher(CbsKafkaConsumerPool kafkaConsumerPool,
                       ExecutorService executorService,
                       @VaccinationETL Function<String, Boolean> vaccinationEtl,
                       @LabResultETL Function<String, Boolean> labResEtl,
                       NasPublisher nasPublisher,
                       DataFetcherConfigProperties dataFetcherConfigProperties) {
        this.executorService = executorService;
        this.nasPublisher = nasPublisher;
        IsAliveServlet.setServiceName(DataFetcherConstants.COMPONENT_NAME);
        this.kafkaConsumerPool = kafkaConsumerPool;
        this.vaccinationEtl = vaccinationEtl;
        this.labResEtl = labResEtl;
        this.processingBeginTimeoutMillis = dataFetcherConfigProperties.getProcessingBeginTimeoutMillis();
        this.processingEndTimeoutMillis = dataFetcherConfigProperties.getProcessingEndTimeoutMillis();
        this.shutdownTimeoutMillis = dataFetcherConfigProperties.getShutdownTimeoutMillis();
        processingPermits = new Semaphore(dataFetcherConfigProperties.getMaxConcurrentProcessings());
    }

    /**
     * 1) Acquire a Kafka consumer from the consumer pool (blocking for x milliseconds if no consumer is available, throwing an exception on timeout). <br/>
     * 2) Use consumer to poll Kafka for up to x number of DataFetcherJobs (blocking for x milliseconds if no jobs available, returning 0 jobs on timeout) <br/>
     * 3) Use the provided ExecutorService to process the jobs concurrently <br/>
     * -- Each job fetches from DDV and/or MIBA, stores result in the database and sends NAS notification. <br/>
     * -- Any exception occurring during processing of an individual job causes the job to be marked as failed and a job-failure notification is sent on NAS <br/>
     */
    public List<DataFetcherJobExecutionResult> processJobs() throws Exception {
        if (shutdownTime != null) {
            log.warn("Not processing new tasks - server is shutting down");
            return Collections.emptyList();
        }
        try {
            if (!processingPermits.tryAcquire(processingBeginTimeoutMillis, TimeUnit.MILLISECONDS)) {
                log.info("No processor permit available within time limit of: " + processingBeginTimeoutMillis + " milliseconds. Returning empty result set to indicate nothing has been processed.");
                return Collections.emptyList();
            }
            return doProcessing();
        } finally {
            processingPermits.release();
        }
    }

    private List<DataFetcherJobExecutionResult> doProcessing() throws Exception {
        log.debug("Polling Kafka for jobs");
        final Set<DataFetcherJob> jobs = kafkaConsumerPool.pollDataFetcherJobs();

        log.debug(String.format("Starting parallel processing of %s jobs", jobs.size()));

        Map<CompletableFuture<DataFetcherJobExecutionResult>, DataFetcherJob> jobTrackingMap = new HashMap<>();
        List<CompletableFuture<DataFetcherJobExecutionResult>> executions = jobs.stream().map(job -> {
            CompletableFuture<DataFetcherJobExecutionResult> future = startExecution(job, executorService);
            jobTrackingMap.put(future, job);
            return future;
        }).collect(Collectors.toList());

        try {
            CompletableFuture.allOf(executions.toArray(new CompletableFuture[0])).get(processingEndTimeoutMillis, TimeUnit.MILLISECONDS);
        } catch (TimeoutException timeoutException) {
            log.warn("Processing did not finish in time! Cancelling currently running job executions!");
            executions.forEach(e -> e.cancel(true)); //cancelling causes the task to complete exceptionally
        } catch (ExecutionException | CancellationException ignored) {
            //exceptions are handled individually for each execution
        }
        handleExceptions(jobTrackingMap, executions);

        List<DataFetcherJobExecutionResult> results = getExecutionResults(jobTrackingMap, executions);
        log.info(String.format("Processed %s jobs", jobs.size()));
        return results;
    }

    private void handleExceptions(Map<CompletableFuture<DataFetcherJobExecutionResult>, DataFetcherJob> jobTrackingMap, List<CompletableFuture<DataFetcherJobExecutionResult>> executions) {
        List<CompletableFuture<DataFetcherJobExecutionResult>> exceptionalExecutions = executions.stream().filter(CompletableFuture::isCompletedExceptionally).collect(Collectors.toList());
        if (!exceptionalExecutions.isEmpty()) {
            log.error(String.format("%s jobs failed with unhandled exceptions! Will send failure notification for each job to NAS", exceptionalExecutions.size())); //note: If the job was interrupted after sending to NAS (UNLIKELY), we will have sent 2 notifications to NAS
            for (CompletableFuture<DataFetcherJobExecutionResult> exceptionalExecution : exceptionalExecutions) {
                //note: These NAS notifications are done sequentially and synchronously on this thread! (that's ok, no need to optimize for abnormal operation)
                exceptionalExecution.handle((nullRes, throwable) -> {
                    final Throwable failure = throwable instanceof CompletionException ? throwable.getCause() : throwable;
                    DataFetcherJob exceptionalJob = jobTrackingMap.get(exceptionalExecution);
                    log.debug(String.format("Job failed with unhandled exception=%s. Job=%s", failure, exceptionalJob));
                    DataFetcherJobExecutionResult resultAfterSending = sendNasFailureNotification(exceptionalJob, failure);
                    exceptionalExecution.obtrudeValue(resultAfterSending); //complete the job with the new result (that has been handled)
                    return null; //not used
                });
            }
        }
    }

    private List<DataFetcherJobExecutionResult> getExecutionResults(Map<CompletableFuture<DataFetcherJobExecutionResult>, DataFetcherJob> jobTrackingMap, List<CompletableFuture<DataFetcherJobExecutionResult>> executions) {
        return executions.stream().map(e -> {
            try {
                DataFetcherJobExecutionResult valueIfAbsent = new DataFetcherJobExecutionResult(jobTrackingMap.get(e), true, "Absent execution result");
                return e.getNow(valueIfAbsent);
            } catch (Exception exception) {
                return new DataFetcherJobExecutionResult(jobTrackingMap.get(e), true, exception.toString());
            }
        }).collect(Collectors.toList());
    }

    private CompletableFuture<DataFetcherJobExecutionResult> startExecution(DataFetcherJob dataFetcherJob, ExecutorService executorService) {
        if (shutdownTime != null && shutdownTime.plusMillis(shutdownTimeoutMillis).isBefore(Instant.now())) {
            return CompletableFuture.completedFuture(new DataFetcherJobExecutionResult(dataFetcherJob, true, CLOSING_SERVER_JOB_FAILURE_MSG));
        }

        if (log.isTraceEnabled())
            log.trace("Starting processing job: " + dataFetcherJob);

        try {
            switch (dataFetcherJob.getType()) {
                case FETCH_VACCINATION_STATUS:
                    return startEtlAndNasNotificationAsync(dataFetcherJob, executorService, vaccinationEtl, VACCINATION_ETL_FAILURE_MSG);
                case FETCH_TEST_STATUS:
                    return startEtlAndNasNotificationAsync(dataFetcherJob, executorService, labResEtl, LAB_RESULT_ETL_FAILURE_MSG);
                case FETCH_ALL:
                    CompletableFuture<DataFetcherJobExecutionResult> vaccinationEtlTask = startEtlAsync(dataFetcherJob, executorService, vaccinationEtl, VACCINATION_ETL_FAILURE_MSG);
                    CompletableFuture<DataFetcherJobExecutionResult> labResEtlTask = startEtlAsync(dataFetcherJob, executorService, labResEtl, LAB_RESULT_ETL_FAILURE_MSG);
                    return vaccinationEtlTask.thenCombine(labResEtlTask, (DataFetcherJobExecutionResult res1, DataFetcherJobExecutionResult res2) -> {
                        final String failureMsg = Stream.of(res1.getFailureMessage(), res2.getFailureMessage())
                            .filter(StringUtils::isNotEmpty)
                            .collect(Collectors.joining(" AND "));
                        return new DataFetcherJobExecutionResult(dataFetcherJob, res1.isFailed() || res2.isFailed(), Strings.emptyToNull(failureMsg));
                    })
                        .thenApply(this::sendNasNotification);
                default:
                    throw new NotImplementedException("Unknown job type: " + dataFetcherJob.getType());
            }
        } catch (Exception exception) {
            log.warn("Exception processing job: " + dataFetcherJob, exception);
            return CompletableFuture.completedFuture(new DataFetcherJobExecutionResult(dataFetcherJob, true, exception.toString()));
        }
    }

    private CompletableFuture<DataFetcherJobExecutionResult> startEtlAsync(DataFetcherJob dataFetcherJob, ExecutorService executorService, Function<String, Boolean> etlFunction, String etlFailureMsg) {
        return CompletableFuture.supplyAsync(() -> performEtl(dataFetcherJob, etlFunction, etlFailureMsg), executorService);
    }

    private CompletableFuture<DataFetcherJobExecutionResult> startEtlAndNasNotificationAsync(DataFetcherJob dataFetcherJob, ExecutorService executorService, Function<String, Boolean> etlFunction, String etlFailureMsg) {
        return CompletableFuture.supplyAsync(() -> {
            DataFetcherJobExecutionResult result = performEtl(dataFetcherJob, etlFunction, etlFailureMsg);
            return sendNasNotification(result);
        }, executorService);
    }

    private DataFetcherJobExecutionResult performEtl(DataFetcherJob dataFetcherJob, Function<String, Boolean> etlFunction, String etlFailureMsg) {
        if (log.isTraceEnabled()) {
            log.trace("Performing ETL on: " + dataFetcherJob);
        }
        boolean etlSuccess = etlFunction.apply(dataFetcherJob.getCpr());
        DataFetcherJobExecutionResult result = new DataFetcherJobExecutionResult(dataFetcherJob, !etlSuccess, etlSuccess ? null : etlFailureMsg);
        if (log.isTraceEnabled()) {
            log.trace("ETL result: " + result);
        }
        return result;
    }

    private DataFetcherJobExecutionResult sendNasNotification(DataFetcherJobExecutionResult executionResult) {
        try {
            final JobStatusNotification nasNotification = new JobStatusNotification();
            nasNotification.setJobId(executionResult.getDataFetcherJob().getJobId());
            nasNotification.setJobStatus(executionResult.isFailed() ? JobStatus.FAILED : JobStatus.OK);
            if (executionResult.isFailed()) {
                nasNotification.setFailedReasonCode(executionResult.getFailureMessage());
            }
            nasPublisher.publish(nasNotification, executionResult.getDataFetcherJob().getCpr());
            if (log.isTraceEnabled()) {
                log.trace("Successfully sent to NAS: " + executionResult);
            }
        } catch (Exception e) {
            log.error("Exception while sending NAS notification! Job has been processed, but no NAS notification has been sent. Job: " + executionResult.getDataFetcherJob(), e);
            executionResult.setFailed(true).setFailureMessage(e.getMessage());
        }
        return executionResult;
    }

    private DataFetcherJobExecutionResult sendNasFailureNotification(DataFetcherJob dataFetcherJob, Throwable exception) {
        DataFetcherJobExecutionResult executionResult = new DataFetcherJobExecutionResult(dataFetcherJob, true, exception.toString());
        return sendNasNotification(executionResult);
    }

    @Override
    public void close() {
        log.info("Closing " + DataFetcher.class.getSimpleName());
        shutdownTime = Instant.now();
        kafkaConsumerPool.close();
        executorService.shutdown();
        try {
            if (!executorService.awaitTermination(shutdownTimeoutMillis, TimeUnit.MILLISECONDS)) {
                executorService.shutdownNow(); //typical implementations will cancel via Thread.interrupt, so any task that fails to respond to interrupts may never terminate.
                log.error("Could not shutdown a running ExecutorService gracefully! Jobs that did not finish in time may have been consumed (committed) in Kafka, but failed job notifications may NOT have been sent to NAS!");
            }
        } catch (InterruptedException e) {
            //interrupted by another thread while waiting for timeout
            executorService.shutdownNow();
            log.error("Could not shutdown a running ExecutorService gracefully! Jobs that did not finish in time may have been consumed (committed) in Kafka, but failed job notifications have maybe not been sent to NAS!");
            Thread.currentThread().interrupt(); //re-interrupt (demanded by SonarLint rule 'java:S2142')
        }
    }
}
