/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import dk.oio.rep.cpr_dk.xml.schemas.core._2006._01._17.SimpleCPRPersonType;
import dk.oio.rep.itst_dk.xml.schemas._2006._01._17.PersonNameStructureType;
import dk.sundhedsdatastyrelsen.cbs.dao.VaccinationRecord;
import dk.sundhedsdatastyrelsen.cbs.dao.impl.jdbc.VaccinationRecordImpl;
import dk.sundhedsdatastyrelsen.cbs.datacache.Writer;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.LabResultETL;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.VaccinationETL;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.ddv.DDVVaccinePasService;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.exceptions.DDVException;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import dk.vaccinationsregister.schemas._2013._12._01.VaccinationStatus;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;

import javax.xml.bind.JAXBException;
import java.time.LocalDateTime;
import java.util.Random;
import java.util.function.BiFunction;
import java.util.function.Function;

@Configuration
public class DataCacheConfiguration {
    private static final Logger log = Logger.getLogger(DataCacheConfiguration.class);
    private final int cacheEntriesTimeToLiveDays;

    public DataCacheConfiguration(@Value("${datafetcher.datacache.records.time-to-live-days:-1}") int cacheEntriesTimeToLiveDays) {
        this.cacheEntriesTimeToLiveDays = cacheEntriesTimeToLiveDays;
    }

    /**
     * Produces an ETL (extract transform load) for vaccinations
     *
     * @param extractor         the extractor, extracts an DDV dto using a cpr number
     * @param dtoTransformer    the dto transformer, transforms the DDV dto to the internal dto
     * @param dtoDaoTransformer the dto to dao transformer, transforms the DDV dto and the internal dto to the dao
     * @param writer            the writer, loads the dao return <code>true</code> when load succeeded else <code>false</code>
     * @return the ETL converter
     */
    @Bean
    @NonNull
    @VaccinationETL
    public Function<String, Boolean> vaccination(
        @NonNull final Function<String, GetVaccinationPassportAsPDFResponseType> extractor,
        @NonNull final Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> dtoTransformer,
        @NonNull final BiFunction<Vaccination, GetVaccinationPassportAsPDFResponseType, VaccinationRecord<Vaccination, GetVaccinationPassportAsPDFResponseType>> dtoDaoTransformer,
        @NonNull final Writer<VaccinationRecord<Vaccination, GetVaccinationPassportAsPDFResponseType>> writer) {
        return cpr -> {
            final GetVaccinationPassportAsPDFResponseType card = extractor.apply(cpr);
            return writer.write(dtoDaoTransformer.apply(dtoTransformer.convert(card), card));
        };
    }

    @Bean
    @NonNull
    public BiFunction<
        Vaccination,
        GetVaccinationPassportAsPDFResponseType,
        VaccinationRecord<Vaccination, GetVaccinationPassportAsPDFResponseType>
        > vaccinationRecordConverter() {
        return (data, source) -> new VaccinationRecordImpl<Vaccination, GetVaccinationPassportAsPDFResponseType>()
            .cpr(data.cpr())
            .creationTime(LocalDateTime.now())
            .traceId("TODO")
            .setData(data)
            .invalidationTime(cacheEntriesTimeToLiveDays > 0 ? LocalDateTime.now().plusDays(cacheEntriesTimeToLiveDays) : null)
            .setSource(source);
    }

    /**
     * DDV extractor using "vaccinepas-service"
     *
     * @param service
     * @return vaccination status for provided cpr
     */
    @Bean
    @NonNull
    public Function<String, GetVaccinationPassportAsPDFResponseType> ddvExtractor(DDVVaccinePasService service) {
        return cpr -> {
            try {
                return service.getVaccinationPassport(cpr);
            } catch (JAXBException e) {
                throw new DDVException(e);
            }
        };
    }

    /**
     * Mock Extractor for DDV
     */
    //@Bean
    @NonNull
    public Function<String, GetVaccinationPassportAsPDFResponseType> mockDdvExtractor() {
        Random random = new Random();
        double meanSleep = 0;
        double variance = 0;
        if (System.getenv().containsKey("MOCK_DDV_SLEEP_MILLIS")) {
            String sleepTime = System.getenv("MOCK_DDV_SLEEP_MILLIS");
            try {
                meanSleep = Long.parseLong(sleepTime);
                variance = meanSleep / 2;
            } catch (NumberFormatException numberFormatException) {
                log.error("Not doing Mock DDV sleep: Could not parse System Property mock.ddv.sleep-millis to a Long: " + sleepTime);
            }
        }

        double finalMeanSleep = meanSleep;
        double finalVariance = variance;
        return cpr -> {
            if (finalMeanSleep > 0) {
                double randomSleep = finalMeanSleep + random.nextGaussian() * finalVariance;
                try {
                    Thread.sleep((long) Math.max(0, randomSleep));
                } catch (InterruptedException interruptedException) {
                    log.error("Mock DDV sleep interrupted");
                    Thread.currentThread().interrupt();
                }
            }

            GetVaccinationPassportAsPDFResponseType response = new GetVaccinationPassportAsPDFResponseType();
            response.setVaccinationAgainst("SARS-CoV-2 (COVID-19)");
            response.setVaccinationStatus(VaccinationStatus.VACCINERET);
            SimpleCPRPersonType pi = new SimpleCPRPersonType();
            pi.setPersonCivilRegistrationIdentifier(cpr);
            PersonNameStructureType pn = new PersonNameStructureType();
            pn.setPersonGivenName("Lars");
            pn.setPersonMiddleName("Linde");
            pn.setPersonSurnameName("Larsen");
            pi.setPersonNameStructure(pn);
            response.setPersonInfo(pi);
            response.setDocumentId("100000007");
            return response;
        };
    }

    @Bean
    @NonNull
    @LabResultETL
    public Function<String, Boolean> mockLabResults() {
        return cpr -> {
            log.debug("Mocked LabResults ETL (nothing saved in database)");
            return true;
        };
    }
}
