/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.config;

import dk.sds.nsp.kafka.clients.consumer.NSPKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaConsumer;
import dk.sundhedsdatastyrelsen.cbs.kafka.CbsKafkaUtils;
import dk.sundhedsdatastyrelsen.cbs.kafka.pooling.CbsKafkaConsumerPool;
import dk.sundhedsdatastyrelsen.cbs.kafka.pooling.PooledClosableObjectFactory;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.pool2.impl.GenericObjectPool;
import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.env.ConfigurableEnvironment;

import java.util.Properties;
import java.util.function.Function;
import java.util.function.Supplier;

@Configuration
public class DataFetcherKafkaConfig {
    private static final Logger log = Logger.getLogger(DataFetcherKafkaConfig.class);
    private final ConfigurableEnvironment env;
    public static final String MSG_KAFKA_TOPIC_PROPERTY_MISSING = "kafka.topic configuration property missing";

    @Autowired
    public DataFetcherKafkaConfig(ConfigurableEnvironment env) {
        this.env = env;
    }

    @Bean
    public CbsKafkaConsumerPool cbsKafkaConsumerPool(@Value("${cbs.kafka.consumer.poll-timeout:100}") int pollTimeout,
                                                     GenericObjectPoolConfig<CbsKafkaConsumer> poolConfig,
                                                     Function<Properties, NSPKafkaConsumer<String, DataFetcherJob>> nspKafkaConsumerConstructor) {
        Properties kafkaConsumerProperties = CbsKafkaUtils.getKafkaProperties("kafka.consumer.", env);

        final String topic = env.getProperty("kafka.topic");
        if (StringUtils.isEmpty(topic)) {
            throw new IllegalArgumentException(MSG_KAFKA_TOPIC_PROPERTY_MISSING);
        }

        Supplier<CbsKafkaConsumer> cbsKafkaConsumerSupplier = () -> {
            Properties instanceProperties = CbsKafkaUtils.getConsumerInstanceProperties(kafkaConsumerProperties);
            return new CbsKafkaConsumer(topic, instanceProperties.getProperty(ConsumerConfig.CLIENT_ID_CONFIG), nspKafkaConsumerConstructor.apply(instanceProperties), pollTimeout);
        };
        log.info("Initializing " + CbsKafkaConsumerPool.class.getSimpleName() + " with poolConfig=: " + poolConfig);
        return new CbsKafkaConsumerPool(new GenericObjectPool<>(new PooledClosableObjectFactory<>(cbsKafkaConsumerSupplier), poolConfig));
    }

}
