/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher.servlet;

import dk.sundhedsdatastyrelsen.cbs.datafetcher.DataFetcher;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.model.DataFetcherJobExecutionResult;
import org.apache.log4j.Logger;
import org.springframework.context.ApplicationContext;

import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;

import static dk.sundhedsdatastyrelsen.cbs.datafetcher.servlet.DataFetcherServletContextListener.SPRING_APPLICATION_CONTEXT;

@WebServlet("/fetch")
public class DataFetcherServlet extends HttpServlet {
    private static final Logger log = Logger.getLogger(DataFetcherServlet.class);

    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response) {
        log.debug("/fetch called");
        ApplicationContext applicationContext = (ApplicationContext) request.getServletContext().getAttribute(SPRING_APPLICATION_CONTEXT);
        DataFetcher dataFetcher = applicationContext.getBean(DataFetcher.class);
        List<DataFetcherJobExecutionResult> processed;
        try {
            processed = dataFetcher.processJobs();
        } catch (Exception e) {
            log.error("Exception occurred during data fetching", e);
            send500Response(response, e);
            return;
        }

        Map<Boolean, List<DataFetcherJobExecutionResult>> map = processed.stream().collect(Collectors.partitioningBy(DataFetcherJobExecutionResult::isFailed));
        List<DataFetcherJobExecutionResult> failures = map.get(true);
        List<DataFetcherJobExecutionResult> successes = map.get(false);

        Map<String, AtomicLong> failureMap = new HashMap<>();
        failures.forEach(failedJob -> failureMap.computeIfAbsent(failedJob.getFailureMessage(), f -> new AtomicLong(0)).incrementAndGet());

        int status = 200;
        if (!failureMap.isEmpty()) {
            log.warn("Processing completed, but failures occurred: " + failureMap);
            status = 500;
        }
        sendResponse(status, response, processed, failures.size(), successes.size(), failureMap);
    }

    private void send500Response(HttpServletResponse response, Exception e) {
        response.setContentType("text/plain; charset=utf-8");
        response.setStatus(500);
        try {
            response.getWriter().write("Exception occurred during data fetching: " + e.getMessage());
        } catch (IOException ioException) {
            log.warn("Could not send response back to client", e);
        }
    }

    private void sendResponse(int status, HttpServletResponse response, List<DataFetcherJobExecutionResult> processed, long failureCount, long successCount, Map<String, AtomicLong> failureMap) {
        response.setContentType("text/plain; charset=utf-8");
        response.setStatus(status);
        try {
            response.getWriter().write(String.format(
                "Processed %s jobs\n" +
                    "SUCCESS: %s\n" +
                    "FAILED : %s\n",
                processed.size(),
                successCount,
                failureCount
            ));

            if (!failureMap.isEmpty()) {
                response.getWriter().write("Failures: " + failureMap);
            }
        } catch (IOException e) {
            log.warn("Could not send response back to client", e);
        }
    }
}
