/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.datafetcher;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.config.DataFetcherConfig;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.config.DataFetcherConfigProperties;
import dk.sundhedsdatastyrelsen.cbs.datafetcher.model.DataFetcherJobExecutionResult;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.kafka.DataFetcherJobBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.DataFetcherJob;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.JobType;
import dk.sundhedsdatastyrelsen.cbs.exceptions.CbsKafkaException;
import dk.sundhedsdatastyrelsen.cbs.kafka.pooling.CbsKafkaConsumerPool;
import dk.sundhedsdatastyrelsen.cbs.nas.facade.NasPublisher;
import dk.sundhedsdatastyrelsen.cbs.test.AssertionTestUtil;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.JobStatusNotification;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;

import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.ForkJoinTask;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Function;
import java.util.stream.Collectors;

import static com.google.common.truth.Truth.assertThat;
import static dk.sundhedsdatastyrelsen.cbs.datafetcher.DataFetcherTest.FailureSource.*;
import static dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils.createRandomDataFetcherJob;
import static dk.sundhedsdatastyrelsen.cbs.test.KafkaTestUtils.createRandomJobs;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.*;

@CbsTest
@SuppressWarnings("java:S5979") /*SonarLint cannot see that @CbsTest has @ExtendWith(MockitoExtension.class)*/
class DataFetcherTest {
    @Mock
    private CbsKafkaConsumerPool cbsKafkaConsumerPool;
    @Mock
    private Function<String, Boolean> vaccinationEtl;
    @Mock
    private Function<String, Boolean> labResEtl;
    @Mock
    private NasPublisher nasPublisher;
    private DataFetcherConfigProperties dataFetcherConfigProperties;
    private final ReentrantLock lock = new ReentrantLock();
    private DataFetcher dataFetcher;
    private final ExecutorService executorService = new DataFetcherConfig().executorService(100);

    enum FailureSource {
        ETL_SELF_CHECK_FAILS,
        ETL_EXCEPTION,
        NAS_EXCEPTION,
    }

    @BeforeEach
    void setUp() {
        dataFetcherConfigProperties = new DataFetcherConfigProperties(
            5,
            10000,
            300000,
            20000);
        dataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);
        lenient().when(vaccinationEtl.apply(anyString())).thenReturn(true);
        lenient().when(labResEtl.apply(anyString())).thenReturn(true);
    }

    @Test
    void test_no_jobs() throws Exception {
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.emptySet());
        assertThat(dataFetcher.processJobs()).isEmpty();
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(vaccinationEtl, labResEtl, nasPublisher);
    }

    @Test
    void test_jobs_processed_successfully() throws Exception {
        final HashSet<DataFetcherJob> jobs = new HashSet<>();
        final int batchSize = 50;
        for (int i = 0; i < batchSize; i++) {
            jobs.add(createRandomDataFetcherJob());
        }
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(jobs);
        final List<DataFetcherJobExecutionResult> results = dataFetcher.processJobs();
        assertEquals(batchSize, results.size());
        assertTrue(results.stream().noneMatch(DataFetcherJobExecutionResult::isFailed));
        verify(nasPublisher, times(batchSize)).publish(any(JobStatusNotification.class), anyString());
    }

    /**
     * Large scope test <br/>
     * - that combines several failure sources in all possible ways: ETL returns false + ETL exception + NAS exception to verify error handling is correct <br/>
     * - that does so in a large job batch size to spotlight any problems with error handling while concurrently processing jobs <br/>
     * <p>
     * Note: do not parameterize this test. It is helpful to test all combinations of failures together in the same job batch
     */
    @Test
    void test_some_jobs_throwing_exceptions_while_others_finish_successfully() throws Exception {
        final String etlExceptionMsg = "ETL EXCEPTION";
        final String nasExceptionMsg = "NAS EXCEPTION";

        final ImmutableSet<FailureSource> failureSource = ImmutableSet.of(ETL_SELF_CHECK_FAILS, ETL_EXCEPTION, NAS_EXCEPTION);
        final Set<Set<FailureSource>> testCases = Sets.powerSet(failureSource) //test all possible combinations of failures
            .stream()
            .filter(tc -> !(tc.contains(ETL_SELF_CHECK_FAILS) && tc.contains(ETL_EXCEPTION))) //remove invalid test case: ETL cannot both throw exception and fail self-check
            .filter(tc -> !(tc.isEmpty())) //remove empty case
            .collect(Collectors.toSet());

        Iterator<Set<FailureSource>> testCasesIterator = testCases.iterator();

        final RandomData randomData = new RandomData(new Random().nextInt());
        final HashSet<DataFetcherJob> jobs = new HashSet<>();
        final int batchSize = 100; //use a high batch size to make concurrent processing occur
        final Map<String, Set<FailureSource>> failingCprs = new HashMap<>();

        //randomly pick the indexes of the jobs that will fail to make sure job order is not a factor in error handling
        final Set[] failJobIndexes = new Set[batchSize];
        randomData.ints(0, batchSize).distinct().limit(testCases.size()).forEach(i -> failJobIndexes[i] = testCasesIterator.next());
        for (int i = 0; i < batchSize; i++) {
            String cpr = randomData.cpr();
            String msgId = UUID.randomUUID().toString();
            if (failJobIndexes[i] != null) {
                Set<FailureSource> testCase = failJobIndexes[i];
                if (testCase.contains(ETL_EXCEPTION)) {
                    failingCprs.computeIfAbsent(cpr, s -> new HashSet<>()).add(ETL_EXCEPTION);
                    when(vaccinationEtl.apply(cpr)).thenThrow(new RuntimeException(etlExceptionMsg));
                    when(labResEtl.apply(cpr)).thenThrow(new RuntimeException(etlExceptionMsg));
                }
                if (testCase.contains(ETL_SELF_CHECK_FAILS)) {
                    failingCprs.computeIfAbsent(cpr, s -> new HashSet<>()).add(ETL_SELF_CHECK_FAILS);
                    when(vaccinationEtl.apply(cpr)).thenReturn(false);
                    when(labResEtl.apply(cpr)).thenReturn(false);
                }
                if (testCase.contains(NAS_EXCEPTION)) {
                    failingCprs.computeIfAbsent(cpr, s -> new HashSet<>()).add(NAS_EXCEPTION);
                    lenient().doThrow(new RuntimeException(nasExceptionMsg)).when(nasPublisher).publish(any(), eq(cpr));
                }
                msgId = testCase.toString();
            }
            DataFetcherJob job = DataFetcherJobBuilder.aDataFetcherJob()
                .withCpr(cpr)
                .withType(JobType.FETCH_ALL)
                .withMsgId(msgId)
                .withJobId(String.valueOf(i))
                .build();
            jobs.add(job);
        }

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(jobs);
        final List<DataFetcherJobExecutionResult> results = dataFetcher.processJobs();

        assertEquals(batchSize, results.size());
        final List<DataFetcherJobExecutionResult> failedJobs = results.stream().filter(DataFetcherJobExecutionResult::isFailed).collect(Collectors.toList());
        assertEquals(testCases.size(), failedJobs.size());
        assertEquals(batchSize - testCases.size(), results.stream().filter(res -> !res.isFailed()).count());

        List<String> failedCprs = failedJobs.stream()
            .map(DataFetcherJobExecutionResult::getDataFetcherJob)
            .map(DataFetcherJob::getCpr)
            .collect(Collectors.toList());
        assertThat(failingCprs.keySet()).containsExactlyElementsIn(failedCprs);

        for (DataFetcherJobExecutionResult failedJob : failedJobs) {
            Set<FailureSource> testCase = failingCprs.get(failedJob.getDataFetcherJob().getCpr());
            final String expectedFailureMsg = testCase.contains(NAS_EXCEPTION) ? nasExceptionMsg : testCase.contains(ETL_EXCEPTION) ? etlExceptionMsg : "ETL failed";
            assertThat(failedJob.getFailureMessage()).endsWith(expectedFailureMsg);
        }

        verify(nasPublisher, times(batchSize)).publish(any(JobStatusNotification.class), anyString());
    }

    @Test
    void test_kafka_exception() throws Exception {
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenThrow(CbsKafkaException.class);
        assertThrows(CbsKafkaException.class, () -> dataFetcher.processJobs());
    }

    @Test
    void test_processing_begin_timeout() throws Exception {
        dataFetcherConfigProperties
            .setMaxConcurrentProcessings(1)
            .setProcessingBeginTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenAnswer(invocation -> {
            lock.lock(); //blocks if lock is already held
            lock.unlock();
            return createRandomJobs(10);
        });

        lock.lock();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(impatientDataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        final List<DataFetcherJobExecutionResult> results = impatientDataFetcher.processJobs(); //should timeout
        assertEquals(0, results.size()); //times out and does not process anything
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(vaccinationEtl, labResEtl, nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_processing_end_timeout() throws Exception {
        final DataFetcherJob job = createRandomDataFetcherJob(JobType.FETCH_VACCINATION_STATUS);
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.singleton(job));
        when(vaccinationEtl.apply(job.getCpr())).thenAnswer(invocation -> {
            lock.lock(); //blocks if lock is already held
            lock.unlock();
            return true;
        });

        dataFetcherConfigProperties.setProcessingEndTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);
        lock.lock();
        final List<DataFetcherJobExecutionResult> results = impatientDataFetcher.processJobs();
        assertEquals(1, results.size());
        final DataFetcherJobExecutionResult result = results.iterator().next();
        assertTrue(result.isFailed());
        assertEquals(job, result.getDataFetcherJob());
        assertThat(result.getFailureMessage()).isNotEmpty();

        verify(vaccinationEtl).apply(job.getCpr());
        verify(nasPublisher).publish(any(JobStatusNotification.class), eq(result.getDataFetcherJob().getCpr()));
        lock.unlock();
    }

    @Test
    void test_shutdown_timeout_when_kafka_hangs() throws Exception {
        dataFetcherConfigProperties.setShutdownTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenAnswer(invocation -> {
            lock.lock(); //blocks if lock is already held
            lock.unlock();
            return createRandomJobs(10);
        });

        lock.lock();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(impatientDataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        impatientDataFetcher.close();
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verifyNoInteractions(vaccinationEtl, labResEtl, nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_shutdown_timeout_when_etl_hangs() throws Exception {
        final DataFetcherJob job = createRandomDataFetcherJob(JobType.FETCH_VACCINATION_STATUS);
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(Collections.singleton(job));

        dataFetcherConfigProperties.setShutdownTimeoutMillis(1);
        final DataFetcher impatientDataFetcher = new DataFetcher(cbsKafkaConsumerPool, executorService, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);

        when(vaccinationEtl.apply(anyString())).thenAnswer(invocation -> {
            lock.lock(); //blocks if lock is already held
            lock.unlock();
            return true;
        });

        lock.lock();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(impatientDataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        impatientDataFetcher.close();
        verify(cbsKafkaConsumerPool).pollDataFetcherJobs();
        verify(vaccinationEtl).apply(anyString());
        verifyNoInteractions(labResEtl, nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_concurrent_invocation_overtaking() throws Exception {
        final DataFetcherJob blockingJob = createRandomDataFetcherJob(JobType.FETCH_ALL);
        final int batchSize = 10;
        //noinspection unchecked
        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(
            Collections.singleton(blockingJob), //first invocation
            createRandomJobs(batchSize) //second invocation
        );

        when(labResEtl.apply(blockingJob.getCpr())).thenAnswer(invocation -> {
            lock.lock(); //blocks if lock is already held
            lock.unlock();
            return true;
        });
        when(vaccinationEtl.apply(blockingJob.getCpr())).thenAnswer(invocation -> {
            lock.lock(); //blocks if lock is already held
            lock.unlock();
            return true;
        });

        lock.lock();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(dataFetcher::processJobs);//will hang
        AssertionTestUtil.waitForAsserted(() -> assertThat(lock.getQueueLength()).isGreaterThan(0));

        final List<DataFetcherJobExecutionResult> results = dataFetcher.processJobs(); //should not hang
        assertEquals(batchSize, results.size());
        assertTrue(results.stream().noneMatch(DataFetcherJobExecutionResult::isFailed));
        verify(nasPublisher, times(batchSize)).publish(any(JobStatusNotification.class), anyString());
        verifyNoMoreInteractions(nasPublisher);

        lock.unlock();
        task.cancel(true);
    }

    @Test
    void test_max_threads_per_processing_is_respected() throws Exception {
        AtomicInteger messagesSentToNas = new AtomicInteger();

        when(cbsKafkaConsumerPool.pollDataFetcherJobs()).thenReturn(createRandomJobs(10));
        doAnswer((invocation -> {
            messagesSentToNas.getAndIncrement();
            lock.lock(); //blocks the worker thread because lock is held by main test thread
            lock.unlock();
            return invocation.getArgument(0);
        })).when(nasPublisher).publish(any(JobStatusNotification.class), anyString());

        final ExecutorService singleThreadedExecutor = new DataFetcherConfig().executorService(1);
        final DataFetcher singleThreadedDataFetcher = new DataFetcher(cbsKafkaConsumerPool, singleThreadedExecutor, vaccinationEtl, labResEtl, nasPublisher, dataFetcherConfigProperties);
        lock.lock();
        final ForkJoinTask<List<DataFetcherJobExecutionResult>> task = ForkJoinPool.commonPool().submit(singleThreadedDataFetcher::processJobs);//will let 1 job through pipeline, and rest will hang
        AssertionTestUtil.ensureAssertedFor500Millis(() -> assertThat(messagesSentToNas.get()).isAtMost(1));

        lock.unlock();
        task.cancel(true);
    }

}
