/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.ddv.converter;

import dk.oio.rep.cpr_dk.xml.schemas.core._2006._01._17.SimpleCPRPersonType;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.vaccinationsregister.schemas._2013._12._01.EffectuationOverviewType;
import dk.vaccinationsregister.schemas._2013._12._01.GetVaccinationPassportAsPDFResponseType;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.convert.converter.Converter;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;

import java.util.Objects;
import java.util.stream.Stream;

import static java.lang.String.format;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.params.provider.Arguments.arguments;

@SpringJUnitConfig(DDVMapperConfiguration.class)
class VaccinationConverterTest {

    @NonNull
    private final Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> converter;
    private static final RandomData RANDOM_DATA = new RandomData("CBS".hashCode());

    @Autowired
    VaccinationConverterTest(
        @NonNull final Converter<GetVaccinationPassportAsPDFResponseType, Vaccination> converter) {
        this.converter = converter;
    }

    @SuppressWarnings("unused") // Idea does not pickup @ParameterizedTest using friendlyName
    @DisplayName("Conversion of Covid19VaccinationCardType to Vaccination")
    @ParameterizedTest(name = "{index}: {1}")
    @MethodSource("cards")
    void convert(@NonNull final GetVaccinationPassportAsPDFResponseType card, @Nullable final String friendly) {
        final Vaccination vaccination = converter.convert(card);
        assertNotNull(vaccination);
        assertEquals(card.getPersonInfo().getPersonCivilRegistrationIdentifier(), vaccination.cpr());
    }

    private static long countEffectuationsWithNonNullDateTime(@NonNull final GetVaccinationPassportAsPDFResponseType card) {
        return card.getEffectuation()
            .stream()
            .map(EffectuationOverviewType::getEffectuatedDateTime)
            .filter(Objects::nonNull)
            .count();
    }

    @NonNull
    private static Stream<Arguments> cards() {
        return generate(RANDOM_DATA::cpr)
            .limit(100)
            .map(Tests::newCard)
            .map(card -> arguments(card, friendly(card)));
    }

    @NonNull
    private static final String FORMAT_STRING = "{" +
        "\"cpr\"=\"%1$s\"," +
        "\"count(effectuatedItems)\"=%2$d," +
        "\"count(completedPlans)\"=%3$d}";

  @NonNull
  private static String friendly(@NonNull final GetVaccinationPassportAsPDFResponseType card) {
    return format(
      FORMAT_STRING,
      card.getPersonInfo().getPersonCivilRegistrationIdentifier(),
      card.getEffectuation().size(),
      countEffectuationsWithNonNullDateTime(card));
  }

  @NonNull
  private static GetVaccinationPassportAsPDFResponseType newCard(@NonNull final String cpr) {
    final GetVaccinationPassportAsPDFResponseType card = new GetVaccinationPassportAsPDFResponseType();
    card.setPersonInfo(new SimpleCPRPersonType());
    card.getPersonInfo().setPersonCivilRegistrationIdentifier(cpr);
    card.getEffectuation().addAll(RANDOM_DATA.list(10, VaccinationConverterTest::newEffectuatedItem));
    return card;
  }

  @NonNull
  private static EffectuationOverviewType newEffectuatedItem() {
    final EffectuationOverviewType itemType = new EffectuationOverviewType();
    if(RANDOM_DATA.nextBoolean())  itemType.setEffectuatedDateTime(RANDOM_DATA.xmlGregorianCalendar());
    return itemType;
  }

}
