/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import dk.sds.nsp.security.Security;
import dk.sds.nsp.security.SecurityContext;
import dk.sds.nsp.security.SecurityContext.User;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.dto.kafka.JobType;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.service.ErrorMessages;
import dk.sundhedsdatastyrelsen.cbs.service.Service;
import dk.sundhedsdatastyrelsen.cbs.service.exception.SecurityRuntimeException;
import dk.sundhedsdatastyrelsen.cbs.spring.Beans;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.CoronaPassPort;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.FutureResult;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.ObjectFactory;
import org.apache.cxf.annotations.SchemaValidation;
import org.apache.log4j.Logger;
import org.springframework.context.ApplicationContext;
import org.springframework.lang.NonNull;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.jws.WebMethod;
import javax.jws.WebResult;
import javax.jws.WebService;
import javax.jws.soap.SOAPBinding;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.xml.bind.annotation.XmlSeeAlso;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import java.util.Optional;

import static dk.sundhedsdatastyrelsen.cbs.dto.kafka.JobType.FETCH_ALL;
import static dk.sundhedsdatastyrelsen.cbs.dto.kafka.JobType.FETCH_TEST_STATUS;
import static dk.sundhedsdatastyrelsen.cbs.dto.kafka.JobType.FETCH_VACCINATION_STATUS;
import static dk.sundhedsdatastyrelsen.cbs.service.servlet.ServiceServletContextListener.SPRING_APPLICATION_CONTEXT;

@SchemaValidation(type = SchemaValidation.SchemaValidationType.BOTH)
@WebService(name = "CoronaPassPort", targetNamespace = "http://sundhedsdatastyrelsen.dk/coronapass/2021/03/coronapass-web-service")
@SOAPBinding(parameterStyle = SOAPBinding.ParameterStyle.BARE)
@XmlSeeAlso({
    ObjectFactory.class
})
public class CoronaPassWebService implements CoronaPassPort {
    private static final Logger LOGGER = Logger.getLogger(CoronaPassWebService.class);

    @Resource
    public WebServiceContext context;

    private Reader<Vaccination, String> vaccinationReader;
    private Reader<LabResults, String> labReader;
    private Service service;
    private MockDataService mockDataService;

    /**
     * @deprecated Used by JAX-WS bean construction
     */
    @Deprecated
    public CoronaPassWebService() {
    }

    @SuppressWarnings("unused")
    @PostConstruct
    public void initialize() {
        initializeSpringDependencies(getSpringContext());
    }

    public CoronaPassWebService(@NonNull final ApplicationContext context) {
        initializeSpringDependencies(context);
    }

    private void initializeSpringDependencies(@NonNull final ApplicationContext context) {
        vaccinationReader = Beans.get(context, Reader.class, Vaccination.class, String.class);
        labReader = Beans.get(context, Reader.class, LabResults.class, String.class);
        mockDataService = context.getBean(MockDataService.class);
        service = context.getBean(Service.class);
    }

    @NonNull
    private ApplicationContext getSpringContext() {
        final ServletContext servletContext = (ServletContext) context.getMessageContext().get(MessageContext.SERVLET_CONTEXT);
        return (ApplicationContext) servletContext.getAttribute(SPRING_APPLICATION_CONTEXT);
    }

    /**
     * @deprecated delete when mock functionality is no longer desired
     */
    @Deprecated
    private Optional<Boolean> getCacheHitHeader() {
        if (context == null) return Optional.empty();
        HttpServletRequest request = (HttpServletRequest) context.getMessageContext().get(MessageContext.SERVLET_REQUEST);
        if (request.getHeader("Xcachehit") == null) return Optional.empty();
        return Optional.of(Boolean.parseBoolean(request.getHeader("Xcachehit")));
    }

    @Override
    @WebMethod(operationName = "GetCoronaPass")
    @WebResult(name = "GetCoronaPassResponse", targetNamespace = "http://sundhedsdatastyrelsen.dk/coronapass/2021/03/coronapass-web-service", partName = "GetCoronaPassResponse")
    public GetCoronaPassResponse getCoronaPass(GetCoronaPassRequest request) {
        SecurityContext securityContext = Security.getSecurityContext();
        final String cpr = checkSecurityTicketGetCpr(securityContext, request);
        final String idwsMessageId = securityContext.getMessage().orElseThrow(RuntimeException::new).getIdentifier().orElseThrow(RuntimeException::new);
        Optional<Boolean> cacheHitHeader = getCacheHitHeader();
        if (cacheHitHeader.isPresent()) {
            if (cacheHitHeader.get()) {
                LOGGER.info("Mocking cache hit (Xcachehit: true http header set)");
                return mockDataService.cacheHit(request.getIdentifier());
            }
            LOGGER.info("Mocking cache miss (Xcachehit: false http header set): Creating DataFetcher job");
            String jobId = service.createDataFetcherJob(request.getIdentifier(), idwsMessageId, FETCH_ALL);
            return cacheMiss(jobId);
        }
        final Vaccination vaccination = vaccinationReader.search(cpr);
        final LabResults labResult = labReader.search(cpr);
        final JobType vaccinationJobType = vaccination == null ? FETCH_VACCINATION_STATUS : null;
        final JobType labResultJobType = labResult == null ? FETCH_TEST_STATUS : null;
        final JobType jobType = JobType.of(vaccinationJobType, labResultJobType);

        if (jobType == null) {
            LOGGER.info("Cache hit");
            return mockDataService.cacheHit(request.getIdentifier());
        } else {
            LOGGER.info("Cache miss");
            final String jobId = service.createDataFetcherJob(cpr, idwsMessageId, jobType);
            return cacheMiss(jobId);
        }
    }

    public GetCoronaPassResponse cacheMiss(String jobId) {
        GetCoronaPassResponse getCoronaPassResponse = new GetCoronaPassResponse();
        FutureResult futureResult = new FutureResult();
        futureResult.setJobId(jobId);
        getCoronaPassResponse.setFutureResult(futureResult);

        return getCoronaPassResponse;
    }

    @NonNull
    protected String checkSecurityTicketGetCpr(
        @NonNull final SecurityContext securityContext,
        @NonNull final GetCoronaPassRequest request) {
        Optional<SecurityContext.Ticket> ticketOptional = securityContext.getTicket();
        if (!ticketOptional.isPresent()) {
            throw new SecurityRuntimeException(ErrorMessages.NO_SECURITY_TICKET);
        }
        SecurityContext.Ticket ticket = ticketOptional.get();
        if (!ticket.isValid()) {
            throw new SecurityRuntimeException(ErrorMessages.INVALID_SECURITY_TICKET);
        }
        final String cprFromTicket = securityContext
            .getActingUser()
            .map(User::getIdentifier)
            .orElseThrow(() -> new SecurityRuntimeException("No user present"));

        final String cpr = request.getIdentifier();

        if (cprFromTicket.equalsIgnoreCase(cpr)) return cpr;

        throw new SecurityRuntimeException(ErrorMessages.NOT_AUTHORIZED_CPR_MISMATCH);
    }

}
