/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.upokecenter.cbor.CBORObject;
import dk.sundhedsdatastyrelsen.cbs.service.webservice.hcertmock.*;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.KeyValuePair;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassList;
import org.springframework.stereotype.Service;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.impl.DefaultDGCEncoder;
import se.digg.dgc.service.impl.DefaultDKEncoder;
import se.digg.dgc.signatures.impl.DefaultDGCSigner;
import se.digg.dgc.signatures.impl.DefaultDKSigner;

import java.io.IOException;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

@Service
public class MockDataService {

    private final KeyStore keystore;
    private final DefaultDGCEncoder euEncoder;
    private final DefaultDKEncoder dkEncoder;

    public MockDataService() {

        try {
            keystore = KeyStore.getInstance("PKCS12");
            keystore.load(this.getClass().getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());

            Certificate certificate = keystore.getCertificate("covid-19 borgerstatus (funktionscertifikat)");
            PrivateKey key = (PrivateKey) keystore.getKey("covid-19 borgerstatus (funktionscertifikat)", "Test1234".toCharArray());

            euEncoder = new DefaultDGCEncoder(new DefaultDGCSigner(key, (X509Certificate) certificate));
            euEncoder.setTransliterateNames(true);
            dkEncoder = new DefaultDKEncoder(new DefaultDKSigner(key, (X509Certificate) certificate));
        } catch (KeyStoreException | IOException | NoSuchAlgorithmException | CertificateException | UnrecoverableKeyException e) {
            throw new RuntimeException("Load of keystore failed:" + e.getMessage(), e);
        }


    }

    private static Map<String, String> jwts = new HashMap<>();

    static {
        jwts.put("DK", "");
        jwts.put("DKPLUS", "");
        jwts.put("EU", "");
    }

    public GetCoronaPassResponse cacheHit(String cpr) {
        try {
            GetCoronaPassResponse getCoronaPassResponse = new GetCoronaPassResponse();
            getCoronaPassResponse.setCoronaPasses(new PassList());

            if (emptyAnswerForCPR(cpr)) return getCoronaPassResponse;

            Person person = person(cpr);
            Instant expiry = MockLogic.expiryInstantForCPR(cpr);



            if(person.dkPass)
            {
                getCoronaPassResponse.getCoronaPasses().getPass().add(getDKPass(cpr, expiry));
                getCoronaPassResponse.getCoronaPasses().getPass().add(getDKPlusPass(cpr, person, expiry));
            }

            if(person.euPass) {

                getEUCertificatePayload(cpr)
                    .stream().forEach(digitalGreenCertificate -> {
                    PassJWT passJWT = new PassJWT();

                    try {
                        passJWT.setJWT(euEncoder.encode(digitalGreenCertificate, expiry));
                    } catch (Exception e) {
                        throw new RuntimeException(e.getMessage(), e);
                    }
                    passJWT.setType("EU");

                    KeyValuePair kv = new KeyValuePair();
                    kv.setKey("eu:keytype");

                    if(!digitalGreenCertificate.getVac().isEmpty())
                        kv.setValue("vac");
                    if(!digitalGreenCertificate.getRec().isEmpty())
                        kv.setValue("rec");
                    if(!digitalGreenCertificate.getTst().isEmpty())
                        kv.setValue("tst");

                    passJWT.getProperties().add(kv);
                    getCoronaPassResponse.getCoronaPasses().getPass().add(passJWT);
                });
            }

            return getCoronaPassResponse;
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    private PassJWT getDKPlusPass(String cpr, Person person, Instant expiry) throws IOException, SignatureException {
        PassJWT passJWT = new PassJWT();
        passJWT.setType("DKPLUS");
        DKPlus dkPlus = new DKPlusDSL().withValidCertificate(cpr, person.first + " " + person.last, "1987-05-04");
        String json = new ObjectMapper().writeValueAsString(dkPlus);
        String encoded = dkEncoder.encodeDkPlus(CBORObject.FromJSONString(json), expiry, UUID.randomUUID().toString());
        passJWT.setJWT(encoded == null ? "" : encoded);
        return passJWT;
    }

    private PassJWT getDKPass(String cpr, Instant expiry) throws IOException, SignatureException {
        PassJWT passJWT = new PassJWT();
        passJWT.setType("DK");
        DK dk = new DKDSL().withValidCertificate(cpr);
        String json = new ObjectMapper().writeValueAsString(dk);
        String encoded = dkEncoder.encodeDk(CBORObject.NewMap(), expiry, UUID.randomUUID().toString());
        passJWT.setJWT(encoded == null ? "" : encoded);
        return passJWT;
    }


    private boolean emptyAnswerForCPR(String cpr) {
        if ("0405870114".equals(cpr) || person(cpr) == null) {
            return true;
        } else {
            return false;
        }
    }

    private static class Person {
        String first;
        String last;
        boolean euPass;
        boolean dkPass;

        public Person(String first, String last, boolean euPass, boolean dkPass) {
            this.first = first;
            this.last = last;
            this.euPass = euPass;
            this.dkPass = dkPass;
        }
    }

    private Person person(String cpr) {
        switch (cpr) {
            case "0405870101":
                return new Person("Anders", "Andersen", true, true);
            case "0405870102":
                return new Person("Bente", "Bendtsen", true, true);
            case "0405870103":
                return new Person("Casper", "Christensen", true, true);
            case "0405870104":
                return new Person("Dorthe", "Drewes", true, true);
            case "0405870105":
                return new Person("Egon", "Eilersen", true, true);
            case "0405870106":
                return new Person("Frederikke", "Frederiksen", true, true);
            case "0405870107":
                return new Person("Gert", "Gorbatjov", true, true);
            case "0405870108":
                return new Person("Helle", "Henriksen", true, true);
            case "0405870109":
                return new Person("Ivan", "Ingolfsen", true, true);
            case "0405870110":
                return new Person("Jytte", "Jensen", true, true);
            case "0405870111":
                return new Person("Klaus", "Kornsted", false, true);
            case "0405870112":
                return new Person("Line", "Larsen", false, true);
            case "0405870113":
                return new Person("Martin", "Monradsen", false, true);
            case "0405870115":
                return new Person("Olga", "Olsen", true, false);
            case "0405870117":
                return new Person("Quang", "Que", true, true);
            case "0405870118":
                return new Person("Rasmus", "Rossen", true, true);
            case "2011451636":
                return new Person("Åge bárður", "Ågesen", true, true);
            case "1704080278":
                return new Person("Tämirä", "Jäkøbsén", true, true);

            default:
                return null;
        }
    }

    public List<DigitalGreenCertificate> getEUCertificatePayload(String cpr) {
        Person name = person(cpr);
        switch (cpr) {
            case "0405870101":
            case "0405870102":
            case "0405870115":
            case "0405870117":
            case "0405870118":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withVaccine()
                    .withTestResult()
                    .withRecovery()
                    .build(cpr);
            case "0405870103":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withVaccine()
                    .withTestResult()
                    .withExpiredRecovery()
                    .build(cpr);
            case "0405870104":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withVaccine()
                    .withExpiredTestResult()
                    .withExpiredRecovery()
                    .build(cpr);
            case "0405870105":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withExpiredVaccine()
                    .withExpiredTestResult()
                    .withExpiredRecovery()
                    .build(cpr);
            case "0405870106":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withExpiredVaccine()
                    .withTestResult()
                    .withExpiredRecovery()
                    .build(cpr);
            case "0405870107":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withExpiredVaccine()
                    .withTestResult()
                    .withRecovery()
                    .build(cpr);
            case "0405870108":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withVaccine()
                    .withExpiredTestResult()
                    .withRecovery()
                    .build(cpr);
            case "0405870109":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withExpiredVaccine()
                    .withExpiredTestResult()
                    .withRecovery()
                    .build(cpr);
            case "0405870110":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withVaccine()
                    .withTestResult()
                    .withRecovery()
                    .expiredBuild();
            case "2011451636":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withTestResult()
                    .withRecovery().build(cpr);
            case "1704080278":
                return new CertificateDSL()
                    .withSubject(name.first, name.last)
                    .withVaccine()
                    .withTestResult()
                    .build(cpr);
            case "0405870111":
            case "0405870112":
            case "0405870113":
                return null;
            default:
                return null;
        }
    }

}
