/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice.hcertmock;


import com.google.common.collect.ImmutableList;
import se.digg.dgc.payload.v1.*;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.UUID;

public class CertificateDSL {

    private final DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd");


    private final ImmutableList.Builder<DigitalGreenCertificate> builder = new ImmutableList.Builder<>();
    private Sub sub;


    public CertificateDSL() {

    }

    public CertificateDSL withVaccine() {

        List<Vac> vacs = new ArrayList<>();

        vacs.add(new Vac()
            .withDis("840539006")
            .withVap("1119349007")
            .withMep("EU/1/20/1528")
            .withAut("ORG-100030215")
            .withSeq(2)
            .withTot(2)
            .withDat(LocalDate.now().minusDays(7))
            .withCou("DK"));
        builder.add(new DigitalGreenCertificate().withDgcid(UUID.randomUUID().toString()).withSub(sub).withVac(vacs).withV("1.0.0"));
        return this;
    }



    public CertificateDSL withTestResult() {

        long sampleDate = LocalDateTime.now().minusDays(1).toEpochSecond(ZoneOffset.UTC);
        long testResult = LocalDateTime.now().minusHours(4).toEpochSecond(ZoneOffset.UTC);

        Tst test = new Tst()
            .withDis("840539006")
            .withTyp("LP6464-4")

            .withTma("BIOSYNEX SWISS SA BIOSYNEX COVID-19 Ag BSS")
            .withOri("258500001")
            .withDtr(Math.toIntExact(sampleDate))
            .withDts(Math.toIntExact(testResult))
            .withRes("1240591000000102")
            .withFac("Region Midtjylland")
            .withCou("DK");
        List<Tst> testList = new ArrayList<>();
        testList.add(test);
        builder.add(new DigitalGreenCertificate().withDgcid(UUID.randomUUID().toString()).withSub(sub).withTst(testList).withV("1.0.0"));
        return this;
    }

    public CertificateDSL withRecovery() {

        Rec recovery = new Rec()
            .withDis("840539006")
            .withDat(LocalDate.now().minusDays(21))
            .withCou("DK");

        List<Rec> recList = new ArrayList<>();
        recList.add(recovery);
        builder.add(new DigitalGreenCertificate().withDgcid(UUID.randomUUID().toString()).withSub(sub).withRec(recList).withV("1.0.0"));
        return this;
    }

    public CertificateDSL withSubject(String givenName, String familyName) {

        this.sub = new Sub()
            .withGn(givenName)
            .withFn(familyName)
            .withDob(LocalDate.now().minusYears(35));
        return this;
    }

    public List<DigitalGreenCertificate> build(String cpr) {
        /*CertificatePayload certificatePayload = getCertificatePayload(LocalDateTime.now(), MockLogic.expiryForCPRLocalDateTime(cpr));
        return certificatePayload;*/
        return builder.build();
    }

    public CertificateDSL withExpiredRecovery() {
        Rec recovery = new Rec()
            .withDis("840539006")
            .withDat(LocalDate.now().minusYears(1))
            .withCou("DK");

        List<Rec> recList = new ArrayList<>();
        recList.add(recovery);

        builder.add(new DigitalGreenCertificate().withDgcid(UUID.randomUUID().toString()).withSub(sub).withRec(recList).withV("1.0.0"));

        return this;
    }

    public CertificateDSL withExpiredTestResult() {

        long sampleDate = LocalDateTime.now().minusDays(5).toEpochSecond(ZoneOffset.UTC);
        long testResult = LocalDateTime.now().minusDays(4).toEpochSecond(ZoneOffset.UTC);

        Tst test = new Tst()
            .withDis("840539006")
            .withTyp("LP6464-4")
            .withTma("BIOSYNEX SWISS SA BIOSYNEX COVID-19 Ag BSS")
            .withOri("258500001")
            .withDtr(Math.toIntExact(sampleDate))
            .withDts(Math.toIntExact(testResult))
            .withRes("1240591000000102")
            .withFac("Region Midtjylland")
            .withCou("DK");
        List<Tst> testList = new ArrayList<>();
        testList.add(test);
        builder.add(new DigitalGreenCertificate().withDgcid(UUID.randomUUID().toString()).withSub(sub).withTst(testList).withV("1.0.0"));

        return this;
    }

    public CertificateDSL withExpiredVaccine() {

        List<Vac> vacs = new ArrayList<>();

        vacs.add(new Vac()
            .withDis("840539006")
            .withVap("1119305005")
            .withMep("EU/1/20/1528")
            .withAut("ORG-100030215")
            .withSeq(2)
            .withTot(2)
            .withDat(LocalDate.now().minusDays(272)) // vaccine udløb
            .withCou("DK"));
        builder.add(new DigitalGreenCertificate().withDgcid(UUID.randomUUID().toString()).withSub(sub).withVac(vacs).withV("1.0.0"));
        return this;
    }

    public List<DigitalGreenCertificate> expiredBuild() {

        return builder.build();

    }
}
