/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.service.webservice;

import dk.sds.nsp.security.DevelopmentSecurityProvider;
import dk.sds.nsp.security.SecurityContext.Ticket;
import dk.sds.nsp.security.SecurityProvider;
import dk.sundhedsdatastyrelsen.cbs.datacache.Reader;
import dk.sundhedsdatastyrelsen.cbs.dto.impl.vaccination.VaccinationBuilder;
import dk.sundhedsdatastyrelsen.cbs.dto.lab.LabResults;
import dk.sundhedsdatastyrelsen.cbs.dto.vaccination.Vaccination;
import dk.sundhedsdatastyrelsen.cbs.test.CbsTest;
import dk.sundhedsdatastyrelsen.cbs.test.ServiceConfiguration;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.PassJWT;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.lang.NonNull;
import org.springframework.test.context.junit.jupiter.SpringJUnitConfig;
import se.digg.dgc.encoding.Base45;
import se.digg.dgc.encoding.Zlib;
import se.digg.dgc.payload.v1.DigitalGreenCertificate;
import se.digg.dgc.service.impl.DefaultDGCDecoder;
import se.digg.dgc.service.impl.MrzParser;
import se.digg.dgc.signatures.CertificateProvider;
import se.digg.dgc.signatures.cose.CoseSign1_Object;
import se.digg.dgc.signatures.cwt.Cwt;
import se.digg.dgc.signatures.impl.DefaultDGCSignatureVerifier;

import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

import static dk.sds.nsp.security.SecurityContext.User;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

@CbsTest
@SpringJUnitConfig({ServiceConfiguration.class})
class CoronaPassWebServiceTest {

    private static Certificate certificate;
    private static DefaultDGCSignatureVerifier verifier;
    private static CertificateProvider certificateProvider;

    private final ApplicationContext applicationContext = new AnnotationConfigApplicationContext("dk.sundhedsdatastyrelsen.cbs.test");
    private final CoronaPassWebService coronaPassWebService = new CoronaPassWebService(applicationContext);

    @NonNull private final Reader<Vaccination, String> vaccinationReader;
    @NonNull private final Reader<LabResults, String> labResultReader;
    @NonNull private final User user;

    @Autowired
    CoronaPassWebServiceTest(
        @NonNull final Reader<Vaccination, String> vaccinationReader,
        @NonNull final Reader<LabResults, String> labResultReader) {
        this.vaccinationReader = vaccinationReader;
        this.labResultReader = labResultReader;
        user = mock(User.class);
        final DevelopmentSecurityProvider securityProvider = (DevelopmentSecurityProvider) SecurityProvider.provider();
        securityProvider.setGlobalTicket(when(mock(Ticket.class).isValid()).thenReturn(true).getMock());
        securityProvider.setGlobalActingUser(user);
    }

    @BeforeAll
    public static void beforeAll() throws KeyStoreException, CertificateException, IOException, NoSuchAlgorithmException {
        KeyStore keystore = KeyStore.getInstance("PKCS12");
        keystore.load(CoronaPassWebServiceTest.class.getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());
        certificate = keystore.getCertificate("covid-19 borgerstatus (funktionscertifikat)");
        verifier = new DefaultDGCSignatureVerifier();
        certificateProvider = (country, kid) -> {
            List<X509Certificate> listOfCertificate = new ArrayList<>();
            listOfCertificate.add((X509Certificate) certificate);
            return listOfCertificate;
        };

    }

    @NonNull
    private GetCoronaPassRequest request(String cpr) {
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(cpr);
        when(user.getIdentifier()).thenReturn(cpr);
        return request;
    }


    @NonNull
    private String cprWithCacheHitOn(@NonNull final String cpr) {
        when(vaccinationReader.search(cpr)).thenReturn(VaccinationBuilder.cpr(cpr).build());
        when(labResultReader.search(cpr)).thenReturn(mock(LabResults.class));
        return cpr;
    }

    @NonNull
    private String cprWithCacheMissOn(@NonNull final String cpr) {
        when(vaccinationReader.search(cpr)).thenReturn(null);
        when(labResultReader.search(cpr)).thenReturn(null);
        return cpr;
    }

    @Test
    void test_getCoronaPass_cache_hit() {
        final String cprPrefix = "040587010";
        for (int i=0; i<10; i++) {
            final String cpr = cprWithCacheHitOn(cprPrefix + 1);
            verifyPass(cpr);
        }
    }

    @Test
    void test_getCoronaPass_empty() {
        String cpr = cprWithCacheHitOn("0405870114");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(0, coronaPass.getCoronaPasses().getPass().size());
    }

    @Test
    void test_getCoronaPassOnlyDK() {
        String cpr = cprWithCacheHitOn("0405870113");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(2, coronaPass.getCoronaPasses().getPass().size());

        PassJWT dkPass = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "DK".equals(p.getType())).findFirst().get();
        assertTrue(dkPass.getJWT().startsWith("DK1:"));
        PassJWT dkPlusPass = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "DKPLUS".equals(p.getType())).findFirst().get();
        assertTrue(dkPlusPass.getJWT().startsWith("DK2:"));
        Cwt dkContent = getPayload(dkPass);
        Cwt dkPlusContent = getPayload(dkPlusPass);

        assertTrue(dkContent.getExpiration().isAfter(Instant.now()));
        assertTrue(dkContent.getIssuedAt().isBefore(Instant.now()));
        assertNotNull(UUID.fromString(new String(dkContent.getCwtId())));

        assertTrue(dkPlusContent.getExpiration().isAfter(Instant.now()));
        assertTrue(dkPlusContent.getIssuedAt().isBefore(Instant.now()));
        assertNotNull(UUID.fromString(new String(dkPlusContent.getCwtId())));
        assertNotNull(dkPlusContent.getClaim("dateOfBirth"));
        assertNotNull(dkPlusContent.getClaim("legalName"));

    }


    @Test
    void test_getCoronaPassOnlyEU() {
        String cpr = cprWithCacheHitOn("0405870115");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(3, coronaPass.getCoronaPasses().getPass().size());
        List<DigitalGreenCertificate> listDGC = payloadToEUpasses(coronaPass.getCoronaPasses().getPass());
        listDGC.forEach(certificate -> {
            String expectedGnt = MrzParser.toMrz(certificate.getSub().getGn(), -1);
            String expectedFnt = MrzParser.toMrz(certificate.getSub().getFn(), -1);
            assertEquals(expectedGnt, certificate.getSub().getGnt());
            assertEquals(expectedFnt, certificate.getSub().getFnt());
        });
    }

    @Test
    void test_getCoronaPassEUMRZnames() {
        String cpr = cprWithCacheHitOn("2011451636");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(4, coronaPass.getCoronaPasses().getPass().size());
        List<DigitalGreenCertificate> listDGC = payloadToEUpasses(coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).collect(Collectors.toList()));
        assertEquals(2, listDGC.size());
        listDGC.forEach(certificate -> {
            String expectedGnt = MrzParser.toMrz(certificate.getSub().getGn(), -1);
            String expectedFnt = MrzParser.toMrz(certificate.getSub().getFn(), -1);
            assertEquals(expectedGnt, certificate.getSub().getGnt());
            assertEquals(expectedFnt, certificate.getSub().getFnt());
        });
    }

    @Test
    void test_getCoronaPass5minutesExpiry() {
        String cpr = cprWithCacheHitOn("0405870118");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(5, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getPayload(eu).getExpiration();
        assertTrue(Instant.now().plus(4, ChronoUnit.MINUTES).isBefore(expDate));
        assertTrue(Instant.now().plus(6, ChronoUnit.MINUTES).isAfter(expDate));
    }

    @Test
    void test_verify_signature() throws Exception {
        String cpr = cprWithCacheHitOn("0405870118");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        DigitalGreenCertificate certificate = getEUPayload(eu.getJWT());
        assertFalse(eu.getProperties().isEmpty());
    }

    @Test
    void test_getCoronaPass15minutesExpiry() {
        String cpr = cprWithCacheHitOn("0405870117");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(5, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getPayload(eu).getExpiration();
        assertTrue(Instant.now().plus(14, ChronoUnit.MINUTES).isBefore(expDate));
        assertTrue(Instant.now().plus(16, ChronoUnit.MINUTES).isAfter(expDate));
    }

    @Test
    void test_getCoronaPass72hourExpiry() {
        String cpr = cprWithCacheHitOn("0405870107");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
        assertEquals(5, coronaPass.getCoronaPasses().getPass().size());
        PassJWT eu = coronaPass.getCoronaPasses().getPass().stream().filter(p -> "EU".equals(p.getType())).findFirst().get();
        Instant expDate = getPayload(eu).getExpiration();
        assertTrue(Instant.now().plus(71, ChronoUnit.HOURS).isBefore(expDate));
        assertTrue(Instant.now().plus(75, ChronoUnit.HOURS).isAfter(expDate));
    }

    private Cwt getPayload(PassJWT eu) {
        return getPayload(eu.getJWT().substring(4));
    }

    private Cwt getPayload(String token) {
        try {
            byte[] cborBytes = Zlib.decompress(Base45.getDecoder().decode(token), true);
            CoseSign1_Object coseObject = new CoseSign1_Object(cborBytes);
            coseObject.verifySignature(certificate.getPublicKey());
            return coseObject.getCwt();
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    public static List<DigitalGreenCertificate> payloadToEUpasses(List<PassJWT> passes)
    {
        return passes.stream().map(PassJWT::getJWT).map(s -> s.substring(4)).map(CoronaPassWebServiceTest::getEUPayload).collect(Collectors.toList());
    }

    public static DigitalGreenCertificate getEUPayload(String token) {
        try {
            return new DefaultDGCDecoder(verifier, certificateProvider).decode(token);
        } catch (Exception e) {
            throw new RuntimeException(e.getMessage(), e);
        }

    }

    private void verifyPass(String cpr) {
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNull(coronaPass.getFutureResult());
    }

    @Test
    void test_getCoronaPass_cache_miss() {
        final String cpr = cprWithCacheMissOn("1234567890");
        GetCoronaPassResponse coronaPass = coronaPassWebService.getCoronaPass(request(cpr));
        assertNotNull(coronaPass);
        assertNotNull(coronaPass.getFutureResult());
        assertNotNull(coronaPass.getFutureResult().getJobId());
        assertNull(coronaPass.getCoronaPasses());
    }}

