/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.test;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.springframework.lang.NonNull;

import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.stream.Stream;

import static java.lang.String.format;
import static java.time.LocalDate.parse;
import static java.time.format.DateTimeFormatter.ofPattern;
import static java.util.stream.Stream.generate;
import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

class RandomDataTest {

  private static final int SEED = "SEED".hashCode();
  private static final int CPR_LENGTH = 10;
  private static final RandomData randomData = new RandomData(SEED);
  private static final DateTimeFormatter CPR_FORMATTER = ofPattern("ddMMyy");


  @DisplayName("cpr numbers")
  @ParameterizedTest(name = "cpr: \"{0}\"")
  @MethodSource("cpr")
  void cpr(@NonNull final String cpr) {
    assertNotNull(cpr);
    assertEquals(CPR_LENGTH, cpr.length(), format("Cpr number %s has invalid length", cpr));
    assertDoesNotThrow(() -> parse(cpr.substring(0, 6), CPR_FORMATTER), "Invalid date");
  }

  @NonNull
  private static Stream<Arguments> cpr() {
    return generate(randomData::cpr).limit(100).map(Arguments::arguments);
  }

  @ParameterizedTest
  @DisplayName("list with positive bound")
  @ValueSource(ints = {1, 2, 4, 8, 16, 32, 64, 128})
  void listValidBound(final int bound) {
    final List<Integer> list = randomData.list(bound, () -> bound);
    assertTrue(bound >= list.size());
    list.forEach((element) -> assertEquals(bound, element));
  }

  @ParameterizedTest
  @DisplayName("list with negative bound throws IllegalArgumentException")
  @ValueSource(ints = {-1, -2, -4, -8, -16, -32, -64, -128})
  void listNegativeBound(final int bound) {
    assertThrows(IllegalArgumentException.class, () -> randomData.list(bound, Object::new));
  }

  @Test
  @DisplayName("list with 0 bound throws IllegalArgumentException")
  void list0Bound() {
    assertThrows(IllegalArgumentException.class, () -> randomData.list(0, Object::new));
  }

  @ParameterizedTest(name = "{index}: max length = {0}")
  @DisplayName("string with positive max length")
  @ValueSource(ints = {1, 2, 4, 8, 16, 32, 64, 128})
  void stringValidBound(final int bound) {
    assertTrue(bound >= randomData.string(bound).length());
  }

  @Test
  @DisplayName("string with max length = 0 throws IllegalArgumentException")
  void string0Bound() {
    assertThrows(IllegalArgumentException.class, () ->randomData.string(0));
  }

  @ParameterizedTest(name = "{index}: max length = {1}")
  @DisplayName("string with negative max length throws IllegalArgumentException")
  @ValueSource(ints = {-1, -2, -4, -8, -16, -32, -64, -128})
  void stringNegativeBound(final int bound) {
    assertThrows(IllegalArgumentException.class, () ->randomData.string(bound));
  }

  @Test
  @DisplayName("XML Gregorian Calendar")
  void xmlCalendar() {
    assertNotNull(randomData.xmlGregorianCalendar());
  }

}
