/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.encoding.impl;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import javax.imageio.ImageIO;

import com.google.zxing.BarcodeFormat;
import com.google.zxing.BinaryBitmap;
import com.google.zxing.DecodeHintType;
import com.google.zxing.MultiFormatReader;
import com.google.zxing.NotFoundException;
import com.google.zxing.Result;
import com.google.zxing.client.j2se.BufferedImageLuminanceSource;
import com.google.zxing.common.HybridBinarizer;

import se.digg.dgc.encoding.BarcodeDecoder;
import se.digg.dgc.encoding.BarcodeException;

/**
 * Default implementation of the {@link BarcodeDecoder} interface using the ZXing library.
 * 
 * @author Martin Lindström (martin@idsec.se)
 * @author Henrik Bengtsson (extern.henrik.bengtsson@digg.se)
 * @author Henric Norlander (extern.henric.norlander@digg.se)
 */
public class DefaultBarcodeDecoder implements BarcodeDecoder {

  /** Decoding hints. */
  private static Map<DecodeHintType, Object> HINTS = new HashMap<>();

  static {
    HINTS.put(DecodeHintType.POSSIBLE_FORMATS, Arrays.asList(BarcodeFormat.QR_CODE, BarcodeFormat.AZTEC));
  }

  /** {@inheritDoc} */
  @Override
  public byte[] decode(final byte[] image) throws BarcodeException {

    BufferedImage bufferedImage = null;
    try (final ByteArrayInputStream imageStream = new ByteArrayInputStream(image)) {
      bufferedImage = ImageIO.read(imageStream);
    }
    catch (IOException e) {
      throw new BarcodeException("Failed to read supplied image bytes into a valid image", e);
    }

    try {
      final MultiFormatReader reader = new MultiFormatReader();
      final Result result = reader.decode(new BinaryBitmap(new HybridBinarizer(new BufferedImageLuminanceSource(bufferedImage))), HINTS);
      if (result.getText() != null) {
        return result.getText().getBytes();
      }
      else {
        throw new BarcodeException("No contents found inside of supplied image");
      }
    }
    catch (final NotFoundException e) {
      throw new BarcodeException("No barcode found inside supplied image", e);
    }
  }

  /** {@inheritDoc} */
  @Override
  public String decodeToString(final byte[] image, final Charset characterSet) throws BarcodeException {
    final byte[] bytes = this.decode(image);
    return new String(bytes, characterSet != null ? characterSet : Charset.defaultCharset());
  }

}
