/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.service;

import com.upokecenter.cbor.CBORObject;

import java.io.IOException;
import java.security.SignatureException;
import java.time.Instant;

/**
 * Service for creating Danish CWTs
 */
public interface DKEncoder {

  /**
   * Based on the DK payload and an expiration time, the method signs it, deflates it, and delivers it in
   * Base45 encoding (with a HCERT header).
   *
   * @param dgc
   *          the contents of the DGC in its CBOR encoding
   * @param expiration
   *          the expiration of the DGC
   * @return the Base45 encoding of the signed DGC
   * @throws IOException
   *           for errors encoding data, for example CBOR related errors
   * @throws SignatureException
   *           errors concerning signing the DGC
   */
  String encodeDk(final CBORObject cborObject, final Instant expiration, String cti) throws IOException, SignatureException;

    /**
     * Based on the DKPLUS payload and an expiration time, the method signs it, deflates it, and delivers it in
     * Base45 encoding (with a HCERT header).
     *
     * @param dgc
     *          the contents of the DGC in its CBOR encoding
     * @param expiration
     *          the expiration of the DGC
     * @return the Base45 encoding of the signed DGC
     * @throws IOException
     *           for errors encoding data, for example CBOR related errors
     * @throws SignatureException
     *           errors concerning signing the DGC
     */
  String encodeDkPlus(final CBORObject cborObject, final Instant expiration,  String cti) throws IOException, SignatureException;

  /**
   * Given the CBOR-encoding of the DGC payload the method creates a CWT and signs it.
   *
   * @param dgc
   *          the DGC payload in its CBOR encoding
   * @param expiration
   *          the expiration of the DGC
   * @return the CBOR encoding of the signed DGC (CWT)
   * @throws IOException
   *           for errors encoding data, for example CBOR related errors
   * @throws SignatureException
   *           errors concerning signing the DGC
   */
  byte[] sign(final CBORObject cborObject, final Instant expiration,  String cti) throws IOException, SignatureException;

}
