/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.signatures;

import java.security.SignatureException;
import java.time.Instant;

/**
 * An interface for a DGC signer.
 * 
 * @author Martin Lindström (martin@idsec.se)
 * @author Henrik Bengtsson (extern.henrik.bengtsson@digg.se)
 * @author Henric Norlander (extern.henric.norlander@digg.se)
 */
public interface DGCSigner {

  /**
   * Creates a CWT including the CBOR encoded DGC payload and signs it.
   * <p>
   * Note: It is the caller's responsibility to ensure that the validity of the issued DGC does not exceed the
   * validity of the signer's certificate (see {@link #getSignerExpiration()}).
   * </p>
   * 
   * @param dgcPayload
   *          the CBOR encoding of the DGC payload
   * @param expiration
   *          the expiration time for the DGC
   * @return the CBOR encoding of the signed CWT holding the DGC payload
   * @throws SignatureException
   *           for signature errors
   */
  byte[] sign(final byte[] dgcPayload, final Instant expiration) throws SignatureException;

  /**
   * Gets the point in time when this signer's certificate expires. For maximum interoperability, the validity of a
   * DGC should not stretch beyond this time.
   * 
   * @return the signer certificate expiration time
   */
  Instant getSignerExpiration();

  /**
   * Gets the ISO-3166 country code of the signer (issuer).
   * 
   * @return the country code of the signer
   */
  String getSignerCountry();

}
