/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.swedenconnect.security.credential;

import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.X509Certificate;

/**
 * Abstract base class for classes implementing the {@link PkiCredential} interface.
 *
 * @author Martin Lindström (martin@idsec.se)
 * @author Stefan Santesson (stefan@idsec.se)
 */
public abstract class AbstractPkiCredential implements PkiCredential {

    /** The private key. */
    private PrivateKey privateKey;

    /** The certificate. */
    private X509Certificate certificate;

    /** The public key. */
    private PublicKey publicKey;

    /** The credential name. */
    private String name;

    /**
     * Default constructor.
     */
    public AbstractPkiCredential() {
    }

    /** {@inheritDoc} */
    @Override
    public PublicKey getPublicKey() {
        return this.certificate != null ? this.certificate.getPublicKey() : this.publicKey;
    }

    /**
     * Assigns the public key of the key pair.
     *
     * @param publicKey
     *          the public key.
     */
    public void setPublicKey(final PublicKey publicKey) {
        if (this.certificate != null) {
            throw new IllegalArgumentException("Cannot assign public key - certificate has already been assigned");
        }
        this.publicKey = publicKey;
    }

    /** {@inheritDoc} */
    @Override
    public X509Certificate getCertificate() {
        return this.certificate;
    }

    /**
     * Assigns the certificate.
     *
     * @param certificate
     *          the certificate
     */
    public void setCertificate(final X509Certificate certificate) {
        if (this.publicKey != null) {
            throw new IllegalArgumentException("Cannot assign certificate - public key has already been assigned");
        }
        this.certificate = certificate;
    }



    /** {@inheritDoc} */
    @Override
    public PrivateKey getPrivateKey() {
        return this.privateKey;
    }

    /**
     * Assigns the private key.
     *
     * @param privateKey
     *          the private key
     */
    public void setPrivateKey(final PrivateKey privateKey) {
        this.privateKey = privateKey;
    }

    /**
     * Gets the name of the credential. If no name has been explicitly assigned, {@link #getDefaultName()} is used.
     */
    @Override
    public String getName() {
        return this.name != null ? this.name : this.getDefaultName();
    }

    /**
     * If the credential {@code name} property is not explicitly assigned using {@link #setName(String)} a name is
     * calculated based on a credential's properties.
     * <p>
     * Implementations must not assume that the object has been correctly initialized.
     * </p>
     *
     * @return the credential name
     */
    protected abstract String getDefaultName();

    /**
     * Assigns the credential name.
     *
     * @param name
     *          the name
     */
    public void setName(final String name) {
        this.name = name;
    }


}
