/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package se.digg.dgc.service.impl;

import com.google.common.collect.Lists;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.x509.X509V1CertificateGenerator;
import org.junit.Assert;
import org.junit.Test;
import org.junit.jupiter.api.Disabled;
import se.digg.dgc.encoding.Barcode;
import se.digg.dgc.encoding.impl.DefaultBarcodeCreator;
import se.digg.dgc.encoding.impl.DefaultBarcodeDecoder;
import se.digg.dgc.payload.v1.*;
import se.digg.dgc.signatures.CertificateProvider;
import se.digg.dgc.signatures.DGCSignatureVerifier;
import se.digg.dgc.signatures.impl.DefaultDGCSignatureVerifier;
import se.digg.dgc.signatures.impl.DefaultDGCSigner;
import se.swedenconnect.security.credential.BasicCredential;
import se.swedenconnect.security.credential.PkiCredential;

import javax.security.auth.x500.X500Principal;
import java.io.IOException;
import java.math.BigInteger;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.time.Duration;
import java.time.Instant;
import java.time.LocalDate;
import java.util.*;

/**
 * Test cases for {@link DefaultDGCBarcodeEncoder}.
 *
 * @author Martin Lindström (martin@idsec.se)
 * @author Henrik Bengtsson (extern.henrik.bengtsson@digg.se)
 * @author Henric Norlander (extern.henric.norlander@digg.se)
 */
public class DefaultDGCBarcodeEncoderDecoderTest {

    private PkiCredential rsa;

    private static final char[] password = "secret".toCharArray();


    public DefaultDGCBarcodeEncoderDecoderTest() throws NoSuchAlgorithmException, NoSuchProviderException, CertificateException, SignatureException, InvalidKeyException {

        Security.insertProviderAt(new BouncyCastleProvider(), 1);
        KeyPairGenerator keyPairGenerator = KeyPairGenerator.getInstance("RSA", "BC");
        keyPairGenerator.initialize(2048, new SecureRandom());

        KeyPair keyPair = keyPairGenerator.generateKeyPair();
        Certificate certificate = generateCertificate(keyPair);
        rsa = new BasicCredential((X509Certificate) certificate, keyPair.getPrivate());
    }

    public static Certificate generateCertificate(KeyPair keyPair) throws CertificateException, NoSuchAlgorithmException, SignatureException, NoSuchProviderException, InvalidKeyException {
        // yesterday
        Date validityBeginDate = new Date(System.currentTimeMillis() - 24 * 60 * 60 * 1000);
        // in 2 years
        Date validityEndDate = new Date(System.currentTimeMillis() + 2 * 365 * 24 * 60 * 60 * 1000);

        X509V1CertificateGenerator certGen = new X509V1CertificateGenerator();
        X500Principal dnName = new X500Principal("CN=John Doe, C=DK");
        certGen.setSerialNumber(BigInteger.valueOf(System.currentTimeMillis()));
        certGen.setSubjectDN(dnName);
        certGen.setIssuerDN(dnName); // use the same
        certGen.setNotBefore(validityBeginDate);
        certGen.setNotAfter(validityEndDate);
        certGen.setPublicKey(keyPair.getPublic());
        certGen.setSignatureAlgorithm("SHA256WithRSAEncryption");
        return certGen.generate(keyPair.getPrivate(), "BC");

    }


    @Test
    @Disabled
    public void decodeToken() throws Exception {

        KeyStore keystore = KeyStore.getInstance("PKCS12");
        keystore.load(this.getClass().getClassLoader().getResourceAsStream("Sundhedsdatastyrelsen - COVID-19 Borgerstatus.p12"), "Test1234".toCharArray());

        Certificate certificate = keystore.getCertificate("covid-19 borgerstatus (funktionscertifikat)");
        PrivateKey key = (PrivateKey) keystore.getKey("covid-19 borgerstatus (funktionscertifikat)", "Test1234".toCharArray());

        DGCSignatureVerifier verifier = new DefaultDGCSignatureVerifier();

        CertificateProvider certificateProvider = (country, kid) -> {
            List<X509Certificate> listOfCertificate = new ArrayList<>();
            listOfCertificate.add((X509Certificate) certificate);
            return listOfCertificate;
        };

        String token = "HC1:NCF3TD2RBFUSPZQEYT2*UE FBUK2RP099O/4C9BRH2S3T5EMVE9EP463EF/8X*G-O9XO9BM9Z0BF8L%N2VZ49:7*5LV 45W0T1LQ8TYD5//C8YIU%A*$S57TNS4HTC/2DBAJDAJCNB-43Q70L.29Z4Z1LF69/2Q*V0JZIX3CXHJBVC8EC400Z1H X40T0X$75-C5-C:NN0KP KLJV0T*8M-JK.GC8CZ8D0YI1OI99B$2T9LTLT3W9TCGKQT3%73%XLE-TN00EIEX02PYENZ6GUADT9H7EH 1W$52Z9POJC2KG7JY1J7UJVTIGZI*RQDZI19JHZII.0F4FQ-B6LF6H5E$BPOJ5OI9YI:8DNFC%PD:NK4WCWVBREJ2YM459KTCMD3QHBRI3DW3POJS7B PIJ1450MWLE JMLII*ED6/D.C8KRDL4O54O4IGUJKJGI2LEYIDOM4R7JJ*7CB9QJPXD0%PQ$76UW6VM9NJPQCCIUQT$4O%2IFT$F30V5L:GQNE$LC25UA+65VCSADM BJ+F0%U4SLM7PUJ0D E58L2DBVUVNSCPHF QBDZR1TFE+O3Q92J45P17CJI.J 4LJ$RO%6A7Q*:AN8O/BPW.RFE1EEUE*P2*RIHN3AIF9MYIGU4BU7IX-49GMGBQ5A6V3D397Q%9OQJA+DL 6D:9:%3K3V+XK/.RB39I7U5VS-YU171MF2MHN3+4.6KC*M9/RA5HTRDI-SJORSPOS6IK4KJLAJ5E%4WO8LKLPSZ4%XQ7 N3:S9K1$T7LNM6DFASR63C4XQHQSCW7B+1F8STYUC7BFETQ/PTWT-0KO16UVUJ.THICEB0YJ1.FGK/85BW+TR8I5M139DNV$U*BS3J0:-QDQ81ORTE7ZECJWCA2TAODJPF$+BRO7SFU2YO7XUZAM0/7R28B-4UUU:KT5 F.TDBY4";
        Assert.assertNotNull(new DefaultDGCDecoder(verifier, certificateProvider).decode(token));
    }

    @Test
    public void testEncodeDecodeBarcode() throws Exception {

        final Instant expire = Instant.now().plus(Duration.ofDays(30));
        final DigitalGreenCertificate dgc = getTestDGC();

        final DefaultDGCBarcodeEncoder encoder = new DefaultDGCBarcodeEncoder(new DefaultDGCSigner(this.rsa), new DefaultBarcodeCreator());

        final Barcode barcode = encoder.encodeToBarcode(dgc, expire);

        final DefaultDGCBarcodeDecoder decoder = new DefaultDGCBarcodeDecoder(
            new DefaultDGCSignatureVerifier(), (x, y) -> Arrays.asList(this.rsa.getCertificate()), new DefaultBarcodeDecoder());

        final DigitalGreenCertificate dgc2 = decoder.decodeBarcode(barcode.getImage());
        Assert.assertEquals(dgc, dgc2);
    }

    private DigitalGreenCertificate getTestDGC() {
        DigitalGreenCertificate dgc = new DigitalGreenCertificate();
        dgc.setV("1.0.0");
        dgc.setDgcid(UUID.randomUUID().toString());
        Sub sub = new Sub();
        sub
            .withGn("Martin")
            .withFn("Lindström")
            .withDob(LocalDate.parse("1969-11-29"));

        final Id pnr = new Id().withT(Id.IdentifierType.NN).withI("196911292932");
        final Id passport = new Id().withT(Id.IdentifierType.PP).withI("56987413").withC("SE");

        sub.withId(Arrays.asList(pnr, passport));

        dgc.setSub(sub);

        Vac vac = new Vac();
        vac
            .withDis("Covid-19")
            .withVap("vap-value")
            .withMep("mep-value")
            .withAut("aut-value")
            .withSeq(Integer.valueOf(1))
            .withTot(Integer.valueOf(2))
            .withDat(LocalDate.parse("2021-04-02"))
            .withCou("SE");

        dgc.setVac(Arrays.asList(vac));

        return dgc;
    }

}
