/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.fluent.spring.jdbc;

import org.junit.jupiter.api.Test;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.lang.NonNull;
import org.springframework.lang.Nullable;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.LocalDateTime;
import java.util.UUID;

import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.Mappers.mapper;
import static dk.sundhedsdatastyrelsen.fluent.spring.jdbc.TableMetaDataBuilder.tableMetaData;
import static java.time.ZoneOffset.UTC;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class MappersTest {

    private static final String CATALOG_NAME = "CATALOG";
    private static final String SCHEMA_NAME = "SCHEMA";
    private static final String TABLE_NAME = "TABLE";
    private static final String TIMESTAMP_LABEL = "TIMESTAMP";
    private static final String LOCAL_DATE_TIME_LABEL = "LOCAL_DATE_TIME";
    private static final String STRING_LABEL = "STRING";
    private static final String INTEGER_LABEL = "INTEGER";
    private static final String UUID_LABEL = "UUID";

    @Test
    void test() throws SQLException {
        final ResultSet resultSet = mock(ResultSet.class);
        final Timestamp timestamp = Timestamp.from(LocalDateTime.now().toInstant(UTC));
        final UUID uuid = UUID.randomUUID();
        final byte[] uuidBytes = UUIDs.toBytes(uuid);
        final String string = "String";

        when(resultSet.getTimestamp(TIMESTAMP_LABEL)).thenReturn(timestamp);
        when(resultSet.getTimestamp(LOCAL_DATE_TIME_LABEL)).thenReturn(timestamp);
        when(resultSet.getString(STRING_LABEL)).thenReturn(string);
        when(resultSet.getInt(INTEGER_LABEL)).thenReturn(0);
        when(resultSet.getBytes(UUID_LABEL)).thenReturn(uuidBytes);
        when(resultSet.wasNull()).thenReturn(true);

        final Mapper<MyDao> mapper = mapper(
            String.class,
            tableMetaData(TABLE_NAME).schemaName(SCHEMA_NAME).catalogName(CATALOG_NAME).build(),
            tableMetaData(TABLE_NAME).schemaName(SCHEMA_NAME).catalogName(CATALOG_NAME).build(),
            MyDao::new
        )
            .addTimestampMapping(TIMESTAMP_LABEL, MyDao::timestamp, MyDao::timestamp)
            .addTimestampToLocalDateTimeMapping
                (LOCAL_DATE_TIME_LABEL, MyDao::localDateTime, MyDao::localDateTime)
            .addBytesMapping(UUID_LABEL, MyDao::uuid, MyDao::uuid)
            .addIntegerMapping(INTEGER_LABEL, MyDao::integer, MyDao::integer)
            .addStringMapping(STRING_LABEL, MyDao::string, MyDao::string)
            .build();

        final MyDao myDao = mapper.mapRow(resultSet, 0);
        assertNotNull(myDao);
        assertEquals(timestamp, myDao.timestamp, TIMESTAMP_LABEL + " mismatch");
        assertEquals(timestamp.toLocalDateTime(), myDao.localDateTime, LOCAL_DATE_TIME_LABEL + " mismatch");
        assertEquals(uuid, myDao.uuid, UUID_LABEL + " mismatch");
        assertEquals(string, myDao.string, STRING_LABEL + " mismatch");
        assertNull(myDao.integer, INTEGER_LABEL + " mismatch");

        final SqlParameterSource source = mapper.parameterSource(myDao);
        assertNotNull(source);
        assertEquals(timestamp, source.getValue(TIMESTAMP_LABEL));
        assertEquals(timestamp, source.getValue(LOCAL_DATE_TIME_LABEL));
        assertArrayEquals(uuidBytes, (byte[]) source.getValue(UUID_LABEL));
        assertEquals(string, source.getValue(STRING_LABEL));
        assertNull(source.getValue(INTEGER_LABEL));
    }

    private static final class MyDao {
        private UUID uuid;
        private String string;
        private Integer integer;
        private Timestamp timestamp;
        private LocalDateTime localDateTime;

        @NonNull
        @SuppressWarnings("UnusedReturnValue")
        public MyDao uuid(@Nullable final UUID uuid) {
            this.uuid = uuid;
            return this;
        }

        @Nullable
        public UUID uuid() {
            return uuid;
        }

        @NonNull
        @SuppressWarnings("UnusedReturnValue")
        public MyDao string(@Nullable final String string) {
            this.string = string;
            return this;
        }

        @Nullable
        public String string() {
            return string;
        }

        @NonNull
        @SuppressWarnings("UnusedReturnValue")
        public MyDao integer(@Nullable final Integer integer) {
            this.integer = integer;
            return this;
        }

        @Nullable
        public Integer integer() {
            return integer;
        }

        @NonNull
        @SuppressWarnings("UnusedReturnValue")
        public MyDao timestamp(@Nullable final Timestamp timestamp) {
            this.timestamp = timestamp;
            return this;
        }

        @Nullable
        public Timestamp timestamp() {
            return timestamp;
        }

        @NonNull
        @SuppressWarnings("UnusedReturnValue")
        public MyDao localDateTime(@Nullable final LocalDateTime localDateTime) {
            this.localDateTime = localDateTime;
            return this;
        }

        @Nullable
        public LocalDateTime localDateTime() {
            return localDateTime;
        }
    }
}
