/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.model.LibertyResponse;
import dk.sosi.seal.vault.CredentialVault;
import dk.sundhedsdatastyrelsen.cbs.configuration.BorgerContext;
import dk.sundhedsdatastyrelsen.cbs.configuration.IntegrationTestConfiguration;
import dk.sundhedsdatastyrelsen.cbs.configuration.RequestContext;
import dk.sundhedsdatastyrelsen.cbs.configuration.XmlDocumentResponseInterceptor;
import dk.sundhedsdatastyrelsen.cbs.test.RandomData;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.CoronaPassPort;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassRequest;
import dk.sundhedsdatastyrelsen.coronapass._2021._03.coronapass_web_service.GetCoronaPassResponse;
import org.apache.log4j.Logger;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.test.context.ContextConfiguration;
import org.springframework.test.context.junit.jupiter.SpringExtension;
import org.w3c.dom.Document;

import javax.xml.bind.JAXB;
import java.io.ByteArrayOutputStream;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotNull;

@ExtendWith(SpringExtension.class)
@ContextConfiguration(classes = {IntegrationTestConfiguration.class})
class CoronaPassITCase {
    private static final Logger log = Logger.getLogger(CoronaPassITCase.class);
    private static final String FMK_AUDIENCE = "https://fmk";
    private static final String IDWS_CPR = "3108732076";
    private static final String IDWS_PID = "PID:9208-2002-2-098076399641";
    private String WS_CPR = "0405870102";

    @Autowired
    CoronaPassPort servicePort;
    @Autowired
    BorgerContext borgerContext;
    @Autowired
    RequestContext requestContext;
    @Autowired
    XmlDocumentResponseInterceptor documentInterceptor;
    @Autowired
    CredentialVault trustVault;
    @Autowired
    OIOIDWSFactory oioIdwsFactory;

    @BeforeEach
    public void setup() {
        borgerContext.clearContext();
        requestContext.clearContext();
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_same_cpr_x_times() {
        for (int i = 0; i < 10; i++) {
            testGetCoronaPasses();
        }
    }

    @Test
    @Disabled("Useful when developing")
    void testGetCoronaPasses_different_cpr_x_times() {
        RandomData randomData = new RandomData(System.currentTimeMillis());
        for (int i = 0; i < 2; i++) {
            WS_CPR = randomData.cpr();
            testGetCoronaPasses();
        }
    }

    @Test
    void testGetCoronaPasses() {
        String messageId = UUID.randomUUID().toString();
        requestContext.setRequestContext(messageId);
        borgerContext.setContext(FMK_AUDIENCE, IDWS_CPR, IDWS_PID, null);
        GetCoronaPassRequest request = new GetCoronaPassRequest();
        request.setIdentifier(IDWS_CPR);

        GetCoronaPassResponse response = servicePort.getCoronaPass(request);
        ByteArrayOutputStream os = new ByteArrayOutputStream();
        JAXB.marshal(response, os);
        log.info("GetCoronaPassResponse=" + os.toString());

        assertNotNull(response);
        Document responseDocument = documentInterceptor.getDocument();
        LibertyResponse libertyResponse = oioIdwsFactory.createResponse(responseDocument);
        assertNotNull(libertyResponse);

        // Notice: in order to validate response signature, we need to add the cert from "oio_idws_response_signing.jks" residing
        // in NSP docker image /pack/wildfly/modules/system/layers/base/dk/sds/nsp/accesshandler/main/ to our trustVault (trust.jks)
        libertyResponse.validateSignatureAndTrust(trustVault);
        assertEquals(messageId, libertyResponse.getRelatesTo());
    }

}
