/*
 * The MIT License
 *
 * Original work sponsored and donated by The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 * Copyright (C) 2021 The Danish Health Data Authority (http://www.sundhedsdatastyrelsen.dk)
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sundhedsdatastyrelsen.cbs.webservice.idws;

import dk.sosi.seal.OIOIDWSFactory;
import dk.sosi.seal.OIOSAMLFactory;
import dk.sosi.seal.model.*;
import dk.sosi.seal.model.dombuilders.OIOBootstrapToIdentityTokenRequestDOMBuilder;
import dk.sosi.seal.pki.Federation;
import dk.sosi.seal.pki.SOSITestFederation;
import dk.sosi.seal.vault.ClasspathCredentialVault;
import dk.sosi.seal.vault.CredentialVault;
import dk.sosi.seal.xml.XmlUtil;
import org.apache.log4j.Logger;
import org.springframework.http.HttpMethod;
import org.springframework.http.RequestEntity;
import org.springframework.http.ResponseEntity;
import org.springframework.web.client.RestTemplate;
import org.w3c.dom.Document;

import java.util.Properties;

public class IdwsTokenHelper {

	public static final String SERVICE_BOOTSTRAP_2_IDWS = "Bst2Idws";

	private static final Logger logger = Logger.getLogger(IdwsTokenHelper.class);

	private CredentialVault credentialVault;
	private Federation federation;
	private String stsUrl;

	public IdwsTokenHelper() {
		this(new ClasspathCredentialVault(new Properties(), "TEST whitelisted SP SOSI alias.jks", "Test1234"));
	}
	
	public IdwsTokenHelper(CredentialVault credentialVault) {
		this(new SOSITestFederation(SignatureUtil.setupCryptoProviderForJVM()), credentialVault, "http://test2.ekstern-test.nspop.dk:8080/sts/services/");
	}
	
	public IdwsTokenHelper(Federation federation, CredentialVault credentialVault, String stsUrl) {
		this.federation = federation;
		this.credentialVault = credentialVault;
		this.stsUrl = stsUrl;
	}
	
	public IdentityToken getIdentityToken(OIOSAMLAssertion bootstrapTokenAssertion, String audience, String pid, String cpr, String procurationCpr) {

		OIOBootstrapToken bootstrapToken = new OIOBootstrapToken(bootstrapTokenAssertion.getDOM().getDocumentElement());
		OIOBootstrapToIdentityTokenRequestDOMBuilder oioBootstrapToIdentityTokenRequestDOMBuilder = new OIOBootstrapToIdentityTokenRequestDOMBuilder();
		oioBootstrapToIdentityTokenRequestDOMBuilder.setSigningVault(credentialVault);
		oioBootstrapToIdentityTokenRequestDOMBuilder.setAudience(audience);
		if (cpr != null) {
			oioBootstrapToIdentityTokenRequestDOMBuilder.setCPRNumberClaim(cpr);
		}
		oioBootstrapToIdentityTokenRequestDOMBuilder.setOIOBootstrapToken(bootstrapToken);
		if (procurationCpr != null) {
			oioBootstrapToIdentityTokenRequestDOMBuilder.setProcurationByCprClaim(procurationCpr);
		}
		Document bootstrapRequest = oioBootstrapToIdentityTokenRequestDOMBuilder.build();
		Document stsBootstrapResponse = callSts(SERVICE_BOOTSTRAP_2_IDWS, bootstrapRequest);

		OIOSAMLFactory oiosamlFactory = new OIOSAMLFactory();
		OIOBootstrapToIdentityTokenResponse bootstrapToIdentityTokenResponse = oiosamlFactory.createOIOBootstrapToIdentityTokenResponseModelBuilder().build(stsBootstrapResponse);

		if(bootstrapToIdentityTokenResponse.getIdentityToken() == null) {
			throw new RuntimeException("Error received from STS.");
		}

		return bootstrapToIdentityTokenResponse.getIdentityToken();
	}

	public OIOIDWSFactory getOioIdwsFactory() {
		return new OIOIDWSFactory(federation, credentialVault);
	}

	public Document callSts(String issueService, Document wstRequest) {
		String stsRequestBody = XmlUtil.node2String(wstRequest, false, true);
		RequestEntity<String> stsRequest = new RequestEntity<String>(stsRequestBody, HttpMethod.POST, null); 
		ResponseEntity<String> stsResult = new RestTemplate().postForEntity(stsUrl+issueService, stsRequest, String.class);
		String responseBody = stsResult.getBody();
		logger.debug("Sts response body: " + responseBody );
		return XmlUtil.readXml(new Properties(), responseBody, false);
	}
}
